import numpy as np
import time
import warnings
from scipy.optimize import minimize_scalar, minimize, Bounds
from scipy.spatial.distance import pdist, cdist, squareform


MACHINE_EPSILON = lambda x: np.abs(x) * np.finfo(np.float64).eps

default_epsilon = 0.1

# FIXME: The p-values for all regressors should be modified. We should be able to compute the upper, lower, and combined p-value

class ConformalPredictionInterval:

    def __init__(self, lower, upper, epsilon):
        self.lower = lower
        self.upper = upper
        self.epsilon = epsilon

    def __contains__(self, y):
        return self.lower <= y <= self.upper
    
    def width(self):
        return self.upper - self.lower
    
    def __repr__(self):
        return repr((self.lower, self.upper))

    def __str__(self):
        return f'({self.lower}, {self.upper})'

class ConformalRegressor:
    '''
    Parent class for the different ridge regressors. Holds common methods
    '''

    def __init__(self, epsilon=default_epsilon):
        self.epsilon = epsilon
    

    '''
    TODO The methods _get_upper and _get_lower could be called with a vector of significance levels,
         say one casual and one highly confident. In pracice, this could be useful when the aim is 
         descision support. This would then have to play nice wiht everything else...
    '''
    def _construct_Gamma(self, lower, upper, epsilon):
        return ConformalPredictionInterval(lower, upper, epsilon)

    @staticmethod
    def _safe_size_check(X):
        if X is None:
            size = 0
        else:
            size = X.shape[0]
        return size
    
    @staticmethod
    def _calculate_p(Alpha, tau=None, c_type='nonconformity'):
        '''
        Method to compute the smoothed p-value, given an array of nonconformity scores where the last element corresponds 
        to the test object, and a random number tau. If tau is None, the non-smoothed p-value is returned.
        '''
        if c_type == 'nonconformity':
            alpha_y = Alpha[-1]
            if tau is not None:
                gt = np.where(Alpha > alpha_y)[0].size
                eq = np.where(Alpha == alpha_y)[0].size
                p_y = (gt + tau * eq) / Alpha.size
            else:
                geq = np.where(Alpha >= alpha_y)[0].size
                p_y = geq / Alpha.size
        elif c_type == 'conformity':
            alpha_y = Alpha[-1]
            if tau is not None:
                lt = np.where(Alpha < alpha_y)[0].size
                eq = np.where(Alpha == alpha_y)[0].size
                p_y = (lt + tau * eq) / Alpha.size
            else:
                leq = np.where(Alpha <= alpha_y)[0].size
                p_y = leq / Alpha.size
        else:
            raise Exception()
        return p_y

    @staticmethod
    def _get_upper(u_dic, epsilon, n):
        try:
            upper = u_dic[int(np.ceil((1 - epsilon)*n))]
        except KeyError:
            upper = np.inf
        return upper


    @staticmethod
    def _get_lower(l_dic, epsilon, n):
        try:
            lower = l_dic[int(np.floor(epsilon*n))]
        except KeyError:
            lower = -np.inf
        return lower


    @staticmethod
    def _vectorised_l_and_u(A, B):
        '''A and B are columns'''
        # Calculate differences
        differences = B[-1] - B
        
        # Create an array to store results
        l = np.empty_like(B, dtype=float)
        u = np.empty_like(B, dtype=float)
        
        # Calculate values where differences are positive
        mask = differences > 0
        l[mask] = (A[mask] - A[-1]) / differences[mask]
        u[mask] = (A[mask] - A[-1]) / differences[mask]
        
        # Assign positive infinity where differences are non-positive
        l[~mask] = -np.inf
        u[~mask] = np.inf
        
        l = np.sort(u, axis=0)[:-1]
        u = np.sort(u, axis=0)[:-1]

        # These are just to avoid messing with the python indexing. Could probably be removed for efficiency
        l_dic = {i+1: val for i, val in enumerate(l)}
        u_dic = {i+1: val for i, val in enumerate(u)}

        return l_dic, u_dic
    

    @staticmethod
    def minimum_training_set(epsilon, bounds='both'):
        '''
        Returns the minimum initial training set size needed to output informative (finite) prediciton sets

        >>> cp = ConformalRegressor()
        >>> cp.minimum_training_set(0.1)
        20

        >>> cp = ConformalRegressor()
        >>> cp.minimum_training_set(0.1, 'upper')
        10

        >>> import numpy as np
        >>> cp = ConformalRegressor()
        >>> cp.minimum_training_set(np.array([0.1, 0.05]))
        40

        '''
        if not hasattr(epsilon, 'shape'):
            # Then it is a scalar
            if bounds == 'both':
                return int(np.ceil(2/epsilon)) 
            else: 
                return int(np.ceil(1/epsilon)) 
        else:
            # Then it is a vector
            if bounds == 'both':
                return int(np.ceil(2/epsilon.min())) 
            else: 
                return int(np.ceil(1/epsilon.min())) 
    

    @staticmethod
    def err(Gamma, y):
        return int(not(y in Gamma))
    

    @staticmethod
    def width(Gamma):
        return Gamma.width()
    

    def learn_many(self, X, y):
        for x1, y1 in zip(X, y):
            self.learn_one(x1,y1)

    
    def process_dataset(self, X, y, epsilon=None, init_train=0, return_results=False):
        if epsilon is None:
            epsilon = self.epsilon

        Err = 0
        Width = 0

        X_train = X[:init_train]
        y_train = y[:init_train]
        X_run = X[init_train:]
        y_run = y[init_train:]

        if return_results:
            res = np.zeros(shape=(y_run.shape[0], 2))
            prediction_sets = {}

        self.learn_initial_training_set(X=X_train, y=y_train)

        time_init = time.time()
        for i, (obj, lab) in enumerate(zip(X_run, y_run)):
            
            # Make prediction
            Gamma = self.predict(obj, epsilon=epsilon) 

            # Check error
            Err += self.err(Gamma, lab)

            # Learn the label
            self.learn_one(obj, lab)
            
            # Width of interval
            width = self.width(Gamma)
            Width += width

            if return_results:
                res[i, 0] = Err
                res[i, 1] = width
                prediction_sets[i] = Gamma

        time_process = time.time() - time_init

        result = {
            'Efficiency': {
                'Average error': Err/self.y.shape[0],
                'Average width': Width/self.y.shape[0],
                'Time': time_process
                }
            }
        if return_results:
            result['Prediction sets'] = prediction_sets
            result['Cummulative Err'] = res[:, 0]
            result['Width'] = res[:, 1]
        
        return result

class ConformalRidgeRegressor(ConformalRegressor):
    '''
    Conformal ridge regression (Algorithm 2.4 in Algorithmic Learning in a Random World)

    Let's create a dataset with noisy evaluations of the function f(x1,x2) = x1+x2:

    >>> import numpy as np
    >>> np.random.seed(31337) # only needed for doctests
    >>> N = 30
    >>> X = np.random.uniform(0, 1, (N, 2))
    >>> y = X.sum(axis=1) + np.random.normal(0, 0.1, N)

    Import the library and create a regressor:

    >>> cp = ConformalRidgeRegressor()

    Learn the whole dataset:

    >>> cp.learn_initial_training_set(X, y)

    Predict an object (output may not be exactly the same, as the dataset
    depends on the random seed):
    >>> interval = cp.predict(np.array([0.5, 0.5]), bounds='both')
    >>> print("(%.2f, %.2f)" % (interval.lower, interval.upper))
    (0.73, 1.23)

    You can of course learn a new data point online:

    >>> cp.learn_one(np.array([0.5, 0.5]), 1.0)

    The prediction set is the closed interval whose boundaries are indicated by the output.

    We can then predict again:

    >>> interval = cp.predict(np.array([2,4]), bounds='both')
    >>> print("(%.2f, %.2f)" % (interval.lower, interval.upper))
    (5.39, 6.33)
    '''
    # TODO: Fix gracefull error handling when the matrix is singular. It should raise an exception, but we could
    #       specify that it can be handled by changing the ridge parameter.

    def __init__(self, a=0, warnings=True, autotune=False, verbose=0, rnd_state=None, studentised=False, epsilon=default_epsilon):
        '''
        The ridge parameter (L2 regularisation) is a.
        Setting autotune=True automatically tunes the ridge parameter using generalized cross validation when learning initial training set.
        '''
        super().__init__(epsilon=epsilon)
        
        self.a = a
        self.X = None
        self.y = None
        self.p = None
        self.Id = None
        self.XTXinv = None

        # Should we raise warnings
        self.warnings = warnings
        # Do we autotune ridge prarmeter on warning
        self.autotune = autotune

        self.verbose = verbose
        self.rnd_gen = np.random.default_rng(rnd_state)

        # Do we use the studentised residuals
        self.studentised = studentised


    def learn_initial_training_set(self, X, y):
        self.X = X
        self.y = y
        self.p = X.shape[1]
        self.Id = np.identity(self.p)
        if self.autotune:
            self._tune_ridge_parameter()
        else:
            self.XTXinv = np.linalg.inv(self.X.T @ self.X + self.a*self.Id)
    

    def learn_one(self, x, y, precomputed=None):
        '''
        Learn a single example. If we have already computed X and XTXinv, use them for update. Then the last row of X is the object with label y.
        >>> cp = ConformalRidgeRegressor()
        >>> cp.learn_one(np.array([1,0]), 1)
        >>> cp.X
        array([[1, 0]])
        >>> cp.y
        array([1])
        '''
        # Learn label y
        if self.y is None:
            self.y = np.array([y])
        else:
            if hasattr(self, 'h'):
                self.y = np.append(self.y, y.reshape(1, self.h), axis=0)
            else:
                self.y = np.append(self.y, y)
    
        if precomputed is not None:
            X = precomputed['X']
            XTXinv = precomputed['XTXinv']

            if X is not None:
                self.X = X
                self.p = self.X.shape[1]
                self.Id = np.identity(self.p)
            
            if XTXinv is not None:
                self.XTXinv = XTXinv
                
            else:
                if self.X.shape[0] == 1:
                    # print(self.X)
                    # print(self.Id)
                    # print(self.a)
                    self.XTXinv = np.linalg.inv(self.X.T @ self.X + self.a * self.Id)
                else:
                    # Update XTX_inv (inverse of Kernel matrix plus regularisation) Use the Sherman-Morrison formula to update the hat matrix
                            #https://en.wikipedia.org/wiki/Sherman%E2%80%93Morrison_formula
                    self.XTXinv -= (self.XTXinv @ np.outer(x, x) @ self.XTXinv) / (1 + x.T @ self.XTXinv @ x)
            
                    # Check the rank
                    if self.warnings:
                        rank_deficient = not(self.check_matrix_rank(self.XTXinv))
                    
        else:
            # Learn object x
            if self.X is None:
                self.X = x.reshape(1,-1)
                self.p = self.X.shape[1]
                self.Id = np.identity(self.p)
            elif self.X.shape[0] == 1:
                self.X = np.append(self.X, x.reshape(1, -1), axis=0)
                self.XTXinv = np.linalg.inv(self.X.T @ self.X + self.a * self.Id)
            else:
                self.X = np.append(self.X, x.reshape(1, -1), axis=0)
                # Update XTX_inv (inverse of Kernel matrix plus regularisation) Use the Sherman-Morrison formula to update the hat matrix
                        #https://en.wikipedia.org/wiki/Sherman%E2%80%93Morrison_formula
                self.XTXinv -= (self.XTXinv @ np.outer(x, x) @ self.XTXinv) / (1 + x.T @ self.XTXinv @ x)
        
                # Check the rank
                if self.warnings:
                    rank_deficient = not(self.check_matrix_rank(self.XTXinv))


    
    def compute_A_and_B_OLD(self, X, XTXinv, y):
        n = X.shape[0]
        # Hat matrix (This block is the time consuming one...)
        H = X @ XTXinv @ X.T
        C = np.identity(n) - H
        A = C @ np.append(y, 0) # Elements of this vector are denoted ai
        B = C @ np.append(np.zeros((n-1,)), 1) # Elements of this vector are denoted bi
        if self.studentised:
            h = H.diagonal()
            A = A / np.sqrt(1 - h)
            B = B / np.sqrt(1 - h)
        # Nonconformity scores are A + yB = y - yhat
        return A, B
    
    def compute_A_and_B(self, X, XTXinv, y):
        """
        Efficient and correct computation of A and B for conformal ridge regression.
        X: (n, d) augmented matrix (last row is test object)
        XTXinv: (d, d) inverse of X.T @ X + a*I for augmented X
        y: (n-1,) training labels (no test label)
        """
        y_ext = np.append(y, 0)  # y with test point (last row) as 0

        # Compute beta using the augmented X and y_ext (just like the old code)
        beta = XTXinv @ X.T @ y_ext  # (d, d) @ (d, n) @ (n,) -> (d,)

        # Fitted values for all points (including test)
        y_hat = X @ beta  # (n, d) @ (d,) -> (n,)

        # Compute hat matrix diagonal for all points using XTXinv (augmented)
        H_diag = np.sum(X @ XTXinv * X, axis=1)  # (n,)

        # Compute last column of H efficiently
        h_col = X @ XTXinv @ X[-1]  # (n, d) @ (d,) -> (n,)

        # A and B for each point
        A = y_ext - y_hat
        B = -h_col
        B[-1] += 1  # e_{-1}[-1] = 1

        if self.studentised:
            A = A / np.sqrt(1 - H_diag + 1e-12)
            B = B / np.sqrt(1 - H_diag + 1e-12)

        return A, B
    

    def predict(self, x, epsilon=None, bounds='both', return_update=False, debug_time=False):
        """
        This function makes a prediction.

        If you start with no training,
        you get a null prediciton between
        -infinity and +infinity.

        >>> cp = ConformalRidgeRegressor()
        >>> cp.predict(np.array([0.506, 0.22, -0.45]), bounds='both')
        (-inf, inf)
        """
        def build_precomputed(X, XTXinv, A, B):
            computed = {
                'X': X, # The updated matrix of objects
                'XTXinv': XTXinv, # The updated kernel matrix
                'A': A,
                'B': B,
            } 
            return computed

        if epsilon is None:
            epsilon = self.epsilon

        if self._safe_size_check(self.X) > 0:

            tic = time.time()
            # Add row to X matrix
            X = np.append(self.X, x.reshape(1, -1), axis=0)
            toc_add_row = time.time() - tic
            n = X.shape[0]
            XTXinv = None

            # Check that the significance level is not too small. If it is, return infinite prediction interval
            if bounds=='both':
                if not (epsilon >= 2/n):
                    if self.warnings:
                        warnings.warn(f'Significance level epsilon is too small for training set. Need at least {int(np.ceil(2/epsilon))} examples. Increase or add more examples')
                    if return_update:
                        return self._construct_Gamma(-np.inf, np.inf, epsilon), build_precomputed(X, XTXinv, None, None)
                    else: 
                        return self._construct_Gamma(-np.inf, np.inf, epsilon)
            else: 
                if not (epsilon >= 1/n):
                    if self.warnings:
                        warnings.warn(f'Significance level epsilon is too small for training set. Need at least {int(np.ceil(1/epsilon))} examples. Increase or add more examples')
                    if return_update:
                        return self._construct_Gamma(-np.inf, np.inf, epsilon), build_precomputed(X, XTXinv, None, None)
                    else: 
                        return self._construct_Gamma(-np.inf, np.inf, epsilon)

            tic = time.time()
            # Update XTX_inv (inverse of Kernel matrix plus regularisation) Use the Sherman-Morrison formula to update the hat matrix
                    #https://en.wikipedia.org/wiki/Sherman%E2%80%93Morrison_formula

            XTXinv = self.XTXinv - (self.XTXinv @ np.outer(x, x) @ self.XTXinv) / (1 + x.T @ self.XTXinv @ x)
            toc_update_XTXinv = time.time() - tic
            
            tic = time.time()
            A, B = self.compute_A_and_B(X, XTXinv, self.y)
            toc_nc = time.time() - tic

            if self.studentised:
                tic = time.time()
                t = (A[:-1] - A[-1]) / (B[-1] - B[:-1])
                t.sort()
                l_dic = {i+1: val for i, val in enumerate(t)}
                u_dic = {i+1: val for i, val in enumerate(t)}
                toc_dics = time.time() - tic
            else:
                tic = time.time()
                l_dic, u_dic = self._vectorised_l_and_u(A, B)
                toc_dics = time.time() - tic

            if bounds=='both':
                lower = self._get_lower(l_dic=l_dic, epsilon=epsilon/2, n=n)
                upper = self._get_upper(u_dic=u_dic, epsilon=epsilon/2, n=n)
            elif bounds=='lower':
                lower = self._get_lower(l_dic=l_dic, epsilon=epsilon, n=n)
                upper = np.inf
            elif bounds=='upper':
                lower = -np.inf
                upper = self._get_upper(u_dic=u_dic, epsilon=epsilon, n=n)
            else: 
                raise Exception

            if debug_time:
                print(f'Add row: {toc_add_row}')
                print(f'Update kernel: {toc_update_XTXinv}')
                print(f'NC scores: {toc_nc}')
                print(f'l and u: {toc_dics}')
                print()
        else:
            # With just one object, and no label, we cannot predict any meaningful interval
            X = x.reshape(1,-1)
            XTXinv = None
            A = None
            B = None

            lower = -np.inf
            upper = np.inf
    
        if return_update:
            return self._construct_Gamma(lower, upper, epsilon), build_precomputed(X, XTXinv, A, B)
        else:
            return self._construct_Gamma(lower, upper, epsilon)

    def compute_p_value(self, x, y, bounds='both', precomputed=None, tau=None, smoothed=True):
        '''
        Computes the smoothed p-value of the example (x, y).
        '''
        if tau is None and smoothed:
            tau = self.rnd_gen.uniform(0, 1)
        if precomputed is not None:
            
            assert np.allclose(x, precomputed['X'][-1])
            A = precomputed['A']
            B = precomputed['B']
        else:
            if self.XTXinv is not None:
                X = np.append(self.X, x.reshape(1, -1), axis=0)
                XTXinv = self.XTXinv - (self.XTXinv @ np.outer(x, x) @ self.XTXinv) / (1 + x.T @ self.XTXinv @ x)
                A, B = self.compute_A_and_B(X, XTXinv, self.y)
            else:
                A, B = None, None 
        
        if A is not None and B is not None:
            if bounds == 'both':
                E = A + y*B
                Alpha = np.zeros_like(A)
                for i, e in enumerate(E):
                    alpha = min((E >= e).sum(), (E<=e).sum())
                    Alpha[i] = alpha
                c_type = 'conformity'
            elif bounds == 'lower':
                Alpha = -(A + y*B)
                c_type = 'nonconformity'
            elif bounds=='upper':
                Alpha = A + y*B
                c_type = 'nonconformity'
            else:
                raise Exception('bounds must be one of "both", "lower", "upper"')

            if smoothed:
                p = self._calculate_p(Alpha, tau, c_type=c_type)
            else: 
                p = self._calculate_p(Alpha, c_type=c_type)
        else:
            if smoothed:
                p = tau
            else:
                p = 1

        return p
    

    def change_ridge_parameter(self, a):
        '''
        Change the ridge parameter
        >>> cp = ConformalRidgeRegressor()
        >>> cp.learn_one(np.array([1,0]), 1)
        >>> cp.change_ridge_parameter(1)
        >>> cp.a
        1
        '''
        self.a = a
        if self.X is not None:
            self.XTXinv = np.linalg.inv(self.X.T @ self.X + self.a * self.Id)


    def _tune_ridge_parameter(self, a0=None):
        '''
        Tune ridge parameter with Generalized cross validation https://pages.stat.wisc.edu/~wahba/stat860public/pdf1/golub.heath.wahba.pdf
        '''
        XTX = self.X.T @ self.X
        n = self.X.shape[0]
        In = np.identity(n)
        def GCV(a):
            try:
                A = self.X @ np.linalg.inv(XTX + a*self.Id) @ self.X.T
                return (1/n)*np.linalg.norm((In - A) @ self.y)**2 / ((1/n)* np.trace(In- A))**2
            except (np.linalg.LinAlgError, ZeroDivisionError):
                return np.inf
        
        # Initial guess
        if a0 is None:
            a0 = 1e-6 # Just a small pertubation to avoid numerical issues

        # Bounds to ensure a >= 0
        res = minimize(GCV, x0=a0, bounds=Bounds(lb=1e-6, keep_feasible=True)) # May be relevant to pass some arguments here, or even use another minimizer.
        a = res.x[0]

        if self.verbose > 0:
            print(f'New ridge parameter: {a}')
        self.change_ridge_parameter(a)


    # TODO
    def prune_training_set(self):
        '''
        Just an idea at the moment, but perhaps we should have some inclusion criteria for examples to only include the informative ones. Could improve accuracy, but also significantly decrease computation time if we have a large dataset.
        '''
        raise NotImplementedError


    def check_matrix_rank(self, M):
        '''
        Check if a matrix has full rank <==> is invertible
        Returns False if matrix is rank deficient
        NOTE In numerical linear algebra it is a bit more subtle. The condition number can tell us more.

        >>> cp = ConformalRidgeRegressor(warnings=False)
        >>> cp.check_matrix_rank(np.array([[1, 0], [1, 0]]))
        False
        >>> cp.check_matrix_rank(np.array([[1, 0], [0, 1]]))
        True
        '''
        if np.linalg.matrix_rank(M) < M.shape[0]:
            if self.warnings:
                warnings.warn(f'The matrix X is rank deficient. Condition number: {np.linalg.cond(M)}. Consider changing the ridge prarmeter')
            return False
        else:
            return True


class KernelConformalRidgeRegressor(ConformalRegressor):

    # TODO Add doctests to methods where applicable

    def __init__(self, kernel, a=0, warnings=True, verbose=0, rnd_state=None, epsilon=default_epsilon):
        '''
        KernelConformalRidgeRegressor requires a kernel. Some common kernels are found in kernels.py, but it is 
        also compatible with (most) kernels from e.g. scikit-learn.
        Custom kernels can also be passed as callable functions.
        '''
        super().__init__(epsilon=epsilon)

        self.a = a
        self.X = None
        self.y = None
        self.p = None
        self.Id = None
        self.K = None
        self.Kinv = None

        self.kernel = kernel

        # Should we raise warnings
        self.warnings = warnings
        
        self.verbose = verbose

        self.rnd_gen = np.random.default_rng(rnd_state)

    
    def learn_initial_training_set(self, X, y):
        self.X = X
        self.y = y
        Id = np.identity(self.X.shape[0])
       
        self.K = self.kernel(self.X)
        self.Kinv = np.linalg.inv(self.K + self.a * Id)


    @staticmethod
    def _update_Kinv(Kinv, k, kappa):
        # print(f'K: {K}')
        # print(f'k: {k}')
        # print(f'kappa: {kappa}')
        d = 1 / (kappa - k.T @ Kinv @ k)
        return np.block([[Kinv + d * Kinv @ k @ k.T @ Kinv, -d * Kinv @ k], [ -d * k.T @ Kinv, d]])


    @staticmethod
    def _update_K(K, k, kappa):
        # print(f'K: {K}')
        # print(f'k: {k}')
        # print(f'kappa: {kappa}')
        return np.block([[K, k], [k.T, kappa]])


    def learn_one(self, x, y, precomputed=None):
        '''
        Learn a single example
        '''
        # Learn label y
        if self.y is None:
            self.y = np.array([y])
        else:
            self.y = np.append(self.y, y)
        
        if precomputed is not None:
            X = precomputed['X']
            K = precomputed['K']
            Kinv = precomputed['Kinv']

            if X is not None:
                self.X = X

            if K is not None and Kinv is not None:
                self.K = K
                self.Kinv = Kinv

            else:
                Id = np.identity(self.X.shape[0])
                self.K = self.kernel(self.X)
                self.Kinv = np.linalg.inv(self.K + self.a * Id)
        
        else:
            # Learn object x
            if self.X is None:
                self.X = x.reshape(1,-1)
                Id = np.identity(self.X.shape[0])
                self.K = self.kernel(self.X)
                self.Kinv = np.linalg.inv(self.K + self.a * Id)
            elif self.X.shape[0] == 1:
                self.X = np.append(self.X, x.reshape(1, -1), axis=0)
                Id = np.identity(self.X.shape[0])
                self.K = self.kernel(self.X)
                self.Kinv = np.linalg.inv(self.K + self.a * Id)
            else:
                k = self.kernel(self.X, x).reshape(-1, 1)
                kappa = self.kernel(x, x)
                self.K = self._update_K(self.K, k, kappa)
                self.Kinv = self._update_Kinv(self.Kinv, k, kappa + self.a)
                self.X = np.append(self.X, x.reshape(1, -1), axis=0)


    @staticmethod
    def compute_A_and_B(X, K, Kinv, y):
        # print(f'X: {X}')
        # print(f'K: {K}')
        # print(f'Kinv: {Kinv}')
        # print(f'y: {y}')
        n = X.shape[0]
        H = Kinv @ K
        C = np.identity(n) - H
        A = C @ np.append(y, 0) # Elements of this vector are denoted ai
        B = C @ np.append(np.zeros((n-1,)), 1) # Elements of this vector are denoted bi
        # Nonconformity scores are A + yB = y - yhat
        return A, B
    
    
    def predict(self, x, epsilon=None, bounds='both', return_update=False, debug_time=False):
        """
        This function makes a prediction.

        If you start with no training,
        you get a null prediciton between
        -infinity and +infinity.

        TODO Add possibility to learn object to save time

        >>> cp = ConformalRidgeRegressor()
        >>> cp.predict(np.array([0.506, 0.22, -0.45]), bounds='both')
        (-inf, inf)
        """
        def build_precomputed(X, K, Kinv, A, B):
            computed = {
                'X': X, # The updated matrix of objects
                'K': K, # The updated kernel matrix
                'Kinv': Kinv,
                'A': A,
                'B': B,
            } 
            return computed

        if epsilon is None:
            epsilon = self.epsilon
        
        if self.X is not None:

            tic = time.time()
            
            # Temporarily update kernel matrix
            k = self.kernel(self.X, x).reshape(-1, 1)
            kappa = self.kernel(x, x)
            K = self._update_K(self.K, k, kappa)
            Kinv = self._update_Kinv(self.Kinv, k, kappa + self.a)

            toc_update_kernel = time.time() - tic

            tic = time.time()
            # Add row to X matrix
            X = np.append(self.X, x.reshape(1, -1), axis=0)
            toc_add_row = time.time() - tic
            n = X.shape[0]

            # Check that the significance level is not too small. If it is, return infinite prediction interval
            if bounds=='both':
                if not (epsilon >= 2/n):
                    if self.warnings:
                        warnings.warn(f'Significance level epsilon is too small for training set. Need at least {int(np.ceil(2/epsilon))} examples. Increase or add more examples')
                    if return_update:
                        return self._construct_Gamma(-np.inf, np.inf, epsilon), build_precomputed(X, K, Kinv, None, None)
                    else: 
                        return self._construct_Gamma(-np.inf, np.inf, epsilon)
            else: 
                if not (epsilon >= 1/n):
                    if self.warnings:
                        warnings.warn(f'Significance level epsilon is too small for training set. Need at least {int(np.ceil(1/epsilon))} examples. Increase or add more examples')
                    if return_update:
                        return self._construct_Gamma(-np.inf, np.inf, epsilon), build_precomputed(X, K, Kinv, None, None)
                    else: 
                        return self._construct_Gamma(-np.inf, np.inf, epsilon)

            
            tic = time.time()
            A, B = self.compute_A_and_B(X, K, Kinv, self.y)
            toc_nc = time.time() - tic

            tic = time.time()
            l_dic, u_dic = self._vectorised_l_and_u(A, B)
            toc_dics = time.time() - tic

            if bounds=='both':
                lower = self._get_lower(l_dic=l_dic, epsilon=epsilon/2, n=n)
                upper = self._get_upper(u_dic=u_dic, epsilon=epsilon/2, n=n)
            elif bounds=='lower':
                lower = self._get_lower(l_dic=l_dic, epsilon=epsilon, n=n)
                upper = np.inf
            elif bounds=='upper':
                lower = -np.inf
                upper = self._get_upper(u_dic=u_dic, epsilon=epsilon, n=n)
            else: 
                raise Exception

            if debug_time:
                print(f'Add row: {toc_add_row}')
                print(f'Update kernel: {toc_update_kernel}')
                print(f'NC scores: {toc_nc}')
                print(f'l and u: {toc_dics}')
                print()
        else:
            # With just one object, and no label, we cannot predict any meaningful interval
            X = x.reshape(1,-1)
            K = None
            Kinv = None
            A = None
            B = None
            
            lower = -np.inf
            upper = np.inf

        if return_update:
            return self._construct_Gamma(lower, upper, epsilon), build_precomputed(X, K, Kinv, A, B)
        else:
            return self._construct_Gamma(lower, upper, epsilon)
        

    def compute_p_value(self, x, y, bounds='both', precomputed=None, tau=None, smoothed=True):
        '''
        Computes the smoothed p-value of the example (x, y).
        '''
        if tau is None and smoothed:
            tau = self.rnd_gen.uniform(0, 1)
        if precomputed is not None:
            
            assert np.allclose(x, precomputed['X'][-1])
            A = precomputed['A']
            B = precomputed['B']

        else:
            if self.Kinv is not None:

                k = self.kernel(self.X, x).reshape(-1, 1)
                kappa = self.kernel(x, x)
                K = self._update_K(self.K, k, kappa)
                Kinv = self._update_Kinv(self.Kinv, k, kappa + self.a)
                X = np.append(self.X, x.reshape(1, -1), axis=0)
                A, B = self.compute_A_and_B(X, K, Kinv, self.y)
            else:
                A, B = None, None 
        
        if A is not None and B is not None:
            if bounds == 'both':
                E = A + y*B
                Alpha = np.zeros_like(A)
                for i, e in enumerate(E):
                    alpha = min((E >= e).sum(), (E<=e).sum())
                    Alpha[i] = alpha
                c_type = 'conformity'
            elif bounds == 'lower':
                Alpha = -(A + y*B)
                c_type = 'nonconformity'
            elif bounds=='upper':
                Alpha = A + y*B
                c_type = 'nonconformity'
            else:
                raise Exception('bounds must be one of "both", "lower", "upper"')

            if smoothed:
                p = self._calculate_p(Alpha, tau, c_type=c_type)
            else: 
                p = self._calculate_p(Alpha, c_type=c_type)
        else:
            if smoothed:
                p = tau
            else:
                p = 1

        return p
    

    def compute_smoothed_p_value(self, x, y, precomputed=None):
        '''
        Computes the smoothed p-value of the example (x, y).
        Smoothed p-values can be used to test the exchangeability assumption.
        '''

        # Inner method to compute the p-value from NC scores
        def calc_p(A, B, y):
            # Nonconformity scores are A + yB = y - yhat
            Alpha = A + y*B
            alpha_y = Alpha[-1]
            gt = np.where(Alpha > alpha_y)[0].shape[0]
            eq = np.where(Alpha == alpha_y)[0].shape[0]
            tau = self.rnd_gen.uniform(0, 1)
            p_y = (gt + tau * eq)/Alpha.shape[0]
            return p_y
        if precomputed is not None:
            A = precomputed['A']
            B = precomputed['B']
            X = precomputed['X']
            K = precomputed['K']
            Kinv = precomputed['Kinv']

            if A is not None and B is not None:
                p_y = calc_p(A, B, y)
            else:
                if Kinv is not None and X is not None and K is not None:

                    A, B = self.compute_A_and_B(X, K, Kinv, self.y)
                    p_y = calc_p(A, B, y)

                else:
                    p_y = self.rnd_gen.uniform(0, 1)
                
        else:
            if self.Kinv is not None:
                
                k = self.kernel(self.X, x).reshape(-1, 1)
                kappa = self.kernel(x, x)
                K = self._update_K(self.K, k, kappa)
                Kinv = self._update_Kinv(self.Kinv, k, kappa + self.a)
                X = np.append(self.X, x.reshape(1, -1), axis=0)

                A, B = self.compute_A_and_B(X, K, Kinv, self.y)
                p_y = calc_p(A, B, y)

            else:
                p_y = self.rnd_gen.uniform(0, 1)
        return p_y


class ConformalNearestNeighboursRegressor(ConformalRegressor):

    def __init__(self, k, distance='euclidean', distance_func=None, aggregation_method='mean', warnings=True, verbose=0, rnd_state=None, epsilon=default_epsilon):
        super().__init__(epsilon=epsilon)
        
        self.k = k
        self.distance = distance
        if distance_func is None:
            self.distance_func = self._standard_distance_func
        else:
            self.distance_func = distance_func
            self.distance = 'custom'

        self.aggregation_method = aggregation_method
        if aggregation_method == 'mean':
            self.agg_func = np.mean
        elif aggregation_method == 'median':
            self.agg_func = np.median

        self.X = None
        self.y = None
        self.D = None

        self.verbose = verbose
        self.rnd_gen = np.random.default_rng(rnd_state)

        self.warnings = warnings


    def _standard_distance_func(self, X, y=None):
        '''
        By default we use scipy to compute distances
        '''
        X = np.atleast_2d(X)
        if y is None:
            dists = squareform(pdist(X, metric=self.distance))
        else:
            y = np.atleast_2d(y)
            dists = cdist(X, y, metric=self.distance)
        return dists
    

    def learn_initial_training_set(self, X, y):
        self.X = X
        self.y = y
        self.D = self.distance_func(X)

    @staticmethod
    def update_distance_matrix(D, d):
        return np.block([[D, d], [d.T, np.array([0])]])

    def learn_one(self, x, y, precomputed=None):
        '''
        precomputed is a dictionary
        {
            'X': X,
            'D': D,
            'A': A,
            'B': B
        }
        '''
        # Learn label y
        if self.y is None:
            self.y = np.array([y])
        else:
            self.y = np.append(self.y, y)

        if precomputed is None:
            # Learn object x
            if self.X is None:
                self.X = x.reshape(1,-1)
                self.D = self.distance_func(self.X)
            else:
                d = self.distance_func(self.X, x)
                self.D = self.update_distance_matrix(self.D, d)
                self.X = np.append(self.X, x.reshape(1, -1), axis=0)
        else:
            self.X = precomputed['X']
            self.D = precomputed['D']
    

    def predict(self, x, epsilon=None, bounds='both', return_update=False, debug_time=False):

        def build_precomputed(X, D, A, B):
            computed = {
                'X': X,
                'D': D,
                'A': A,
                'B': B
            }
            return computed

        if epsilon is None:
            epsilon = self.epsilon
        
        if self._safe_size_check(self.X) > self.k:

            # Temporarily update distance matrix
            tic = time.time()
            d = self.distance_func(self.X, x)
            D = self.update_distance_matrix(self.D, d)
            toc_update_D = time.time() - tic

            tic = time.time()
            # Add row to X matrix
            X = np.append(self.X, x.reshape(1, -1), axis=0)
            toc_add_row = time.time() - tic
            n = X.shape[0]

            # Check that the significance level is not too small. If it is, return infinite prediction interval
            if bounds=='both':
                if not (epsilon >= 2/n):
                    if self.warnings:
                        warnings.warn(f'Significance level epsilon is too small for training set. Need at least {int(np.ceil(2/epsilon))} examples. Increase or add more examples')
                    if return_update:
                        return self._construct_Gamma(-np.inf, np.inf, epsilon), build_precomputed(X, D, None, None)
                    else: 
                        return self._construct_Gamma(-np.inf, np.inf, epsilon)
            else: 
                if not (epsilon >= 1/n):
                    if self.warnings:
                        warnings.warn(f'Significance level epsilon is too small for training set. Need at least {int(np.ceil(1/epsilon))} examples. Increase or add more examples')
                    if return_update:
                        return self._construct_Gamma(-np.inf, np.inf, epsilon), build_precomputed(X, D, None, None)
                    else: 
                        return self._construct_Gamma(-np.inf, np.inf, epsilon)
            
            tic = time.time()
            A, B = self.compute_A_and_B(D, self.y, self.k)
            toc_nc = time.time() - tic

            tic = time.time()
            l_dic, u_dic = self._vectorised_l_and_u(A, B)
            toc_dics = time.time() - tic

            if bounds=='both':
                lower = self._get_lower(l_dic=l_dic, epsilon=epsilon/2, n=n)
                upper = self._get_upper(u_dic=u_dic, epsilon=epsilon/2, n=n)
            elif bounds=='lower':
                lower = self._get_lower(l_dic=l_dic, epsilon=epsilon, n=n)
                upper = np.inf
            elif bounds=='upper':
                lower = -np.inf
                upper = self._get_upper(u_dic=u_dic, epsilon=epsilon, n=n)
            else: 
                raise Exception

            if debug_time:
                print(f'Add row: {toc_add_row}')
                print(f'Update D: {toc_update_D}')
                print(f'NC scores: {toc_nc}')
                print(f'l and u: {toc_dics}')
                print()
        else:
            # With just one object, and no label, we cannot predict any meaningful interval
            lower = -np.inf
            upper = np.inf
            X = x.reshape(1,-1)
            D = self.distance_func(X)
            if self._safe_size_check(self.X) > 0:
                A, B = self.compute_A_and_B(D, self.y, self.k)
            else:
                A, B = None, None
        
        if return_update:
            return self._construct_Gamma(lower, upper, epsilon), build_precomputed(X, D, A, B)
        else:
            return self._construct_Gamma(lower, upper, epsilon)


    # FIXME This is very wrong!!
    def compute_A_and_B(self, D, y, k):
        y = np.append(y, 0)
        n = D.shape[0] - 1 
        A = np.zeros(n + 1)
        B = np.zeros(n + 1)

        k_nearest = D.argsort(axis=0)[1:k+1]

        for i, col in enumerate(k_nearest.T): # Transpose to iterate over columns
            if i < n:
                A[i] = y[i] - self.agg_func(y[col])
                if n in col:
                    B[i] = -1/k
                else:
                    B[i] = 0
            else:
                A[i] = - self.agg_func(y[col])
                B[i] = 1
        print(A)
        print(B)
        return A, B
    

    @staticmethod
    def _calculate_p(Alpha, tau=None):
        '''
        Method to compute the smoothed p-value, given an array of nonconformity scores where the last element corresponds 
        to the test object, and a random number tau. If tau is None, the non-smoothed p-value is returned.
        '''
        alpha_y = Alpha[-1]
        if tau is not None:
            gt = np.where(Alpha > alpha_y)[0].size
            eq = np.where(Alpha == alpha_y)[0].size
            p_y = (gt + tau * eq) / Alpha.size
        else:
            geq = np.where(Alpha >= alpha_y)[0].size
            p_y = geq / Alpha.size
        return p_y

    def compute_p_value(self, x, y, bounds='both', precomputed=None, tau=None, smoothed=True):
        '''
        Computes the smoothed p-value of the example (x, y).
        '''
        if tau is None and smoothed:
            tau = self.rnd_gen.uniform(0, 1)
        if precomputed is not None:
            
            assert np.allclose(x, precomputed['X'][-1])
            D = precomputed['D']
            A, B = self.compute_A_and_B(D, self.y, self.k)
        else:
            if self._safe_size_check(self.X) > self.k:
                d = self.distance_func(self.X, x)
                D = self.update_distance_matrix(self.D, d)
                A, B = self.compute_A_and_B(D, self.y, self.k)
            else:
                X = x.reshape(1,-1)
                D = self.distance_func(X)
                if self._safe_size_check(self.X) > 0:
                    self.compute_A_and_B(D, self.y, self.k)
                else:
                    A, B = None, None 
        
        if A is not None and B is not None:
            if bounds == 'both':
                E = A + y*B
                Alpha = np.zeros_like(A)
                for i, e in enumerate(E):
                    alpha = min((E >= e).sum(), (E<=e).sum())
                    Alpha[i] = alpha
            elif bounds == 'lower':
                Alpha = -(A + y*B)
            elif bounds=='upper':
                Alpha = A + y*B
            else:
                raise Exception('bounds must be one of "both", "lower", "upper"')

            if smoothed:
                p = self._calculate_p(Alpha, tau)
            else: 
                p = self._calculate_p(Alpha)
        else:
            if smoothed:
                p = tau
            else:
                p = 1

        return p
    

    # def compute_smoothed_p_value(self, x, y, precomputed=None):
    #     '''
    #     Computes the smoothed p-value of the example (x, y).
    #     Smoothed p-values can be used to test the exchangeability assumption.
    #     If X and XTXinv are passed, x must be the last row of X.
    #     '''

    #     # Inner method to compute the p-value from NC scores
    #     def calc_p(A, B, y):
    #         # Nonconformity scores are A + yB = y - yhat
    #         Alpha = A + y*B
    #         alpha_y = Alpha[-1]
    #         gt = np.where(Alpha < alpha_y)[0].shape[0]
    #         eq = np.where(Alpha == alpha_y)[0].shape[0]
    #         tau = self.rnd_gen.uniform(0, 1)
    #         p_y = (gt + tau * eq)/Alpha.shape[0]
    #         return p_y
        
    #     if precomputed is not None:
    #         A = precomputed['A']
    #         B = precomputed['B']
    #         if A is not None and B is not None:
    #             p_y = calc_p(A, B, y)
    #         else:
    #             X = precomputed['X']
    #             if self.X is not None:
    #                 assert np.allclose(x, X[-1])
    #                 D = precomputed['D']
    #                 A, B = self.compute_A_and_B(D, self.y, self.k)
    #                 p_y = calc_p(A, B, y)
    #             else:
    #                 p_y = self.rnd_gen.uniform(0, 1)
    #     else:
    #         if self.X is not None:
    #             d = self.distance_func(self.X, x)
    #             D = self.update_distance_matrix(self.D, d)
    #             A, B = self.compute_A_and_B(D, self.y, self.k)
    #             p_y = calc_p(A, B, y)
    #         else:
    #             p_y = self.rnd_gen.uniform(0, 1)
    #     return p_y
    

if __name__ == "__main__":
    import doctest
    import sys

    (failures, _) = doctest.testmod(optionflags=doctest.REPORT_ONLY_FIRST_FAILURE) # XXX: DO NOT COMMIT ME!!!!!!!!!!!!!!!!!!!!!!
    if failures:
        sys.exit(1)

        

    def test_equivalence_of_A_and_B():
        """
        Test that compute_A_and_B_OLD and compute_A_and_B return (almost) identical results.
        Test on the California Housing dataset, rather than synthetic data.
        """
        # np.random.seed(42)
        # n, d = 10, 3
        # X = np.random.randn(n, d)
        # y = np.random.randn(n-1)

        # Load California housing dataset
        from sklearn.datasets import fetch_california_housing
        data = fetch_california_housing()
        X = np.array(data.data)
        y = np.array(data.target)
        a=0

        # Create a dummy regressor instance
        cp = ConformalRidgeRegressor(a=a)
        cp.learn_initial_training_set(X=X[:-1], y=y[:-1])

        x = X[-1]
        X_internal = np.append(cp.X, x.reshape(1, -1), axis=0)
        XTXinv_internal = cp.XTXinv - (cp.XTXinv @ np.outer(x, x) @ cp.XTXinv) / (1 + x.T @ cp.XTXinv @ x)

        # OLD
        A_old, B_old = cp.compute_A_and_B_OLD(X_internal, XTXinv_internal, cp.y)
        # NEW
        A_new, B_new = cp.compute_A_and_B(X_internal, XTXinv_internal, cp.y)

        print("A_old:", A_old)
        print("A_new:", A_new)
        print("B_old:", B_old)
        print("B_new:", B_new)

        assert np.allclose(A_old, A_new, atol=1e-8), "A vectors differ!"
        assert np.allclose(B_old, B_new, atol=1e-8), "B vectors differ!"
        print("Test passed: compute_A_and_B_OLD and compute_A_and_B are equivalent.")

    # Uncomment to run the test
    test_equivalence_of_A_and_B()


# This file should contain the code to keep track of evaluation metrics
import numpy as np
import matplotlib.pyplot as plt

# NOTE: Below are some efficiency criteria, some of which are proper scoring rules.
#       What I have to do now, is to create one or more Evaluation classes, which, in their
#       construction create metric classes, and keeps track of the statistics as we go on.

import numpy as np

# This is a bit messy...
class Evaluation:
    """
    Class to evaluate and keep track of various metrics.
    """

    def __init__(self, metrics=None, **kwargs):
        """
        Initialize the Evaluation class with specified metrics.

        Parameters:
        metrics (list): List of metric names to be tracked.
        kwargs (dict): Additional metric classes to be tracked.
        """
        self.metrics = {}
        self.results = {}

        self.n = 0

        available_metrics = {
            "oe": OE,
            "of": OF,
            "err": Err,
            "width": Width,
            "winkler": WinklerScore,
            "crps": CRPS,
        }

        if metrics is not None:
            for metric_name in metrics:
                if metric_name.lower() in available_metrics:
                    self._add_metric_instance(metric_name.lower(), available_metrics[metric_name.lower()])
                else:
                    raise ValueError(f"Unknown metric: {metric_name}")

        for metric_name, metric_class in kwargs.items():
            if isinstance(metric_class, type) and issubclass(metric_class, (OE, OF, Err, Width, WinklerScore, CRPS)):
                self._add_metric_instance(metric_name.lower(), metric_class)
            elif metric_name.lower() in available_metrics:
                self._add_metric_instance(metric_name.lower(), available_metrics[metric_name.lower()])
            else:
                raise ValueError(f"Unknown metric: {metric_name}")

    def _add_metric_instance(self, name, metric_class):
        """
        Add an instance of a metric class to the evaluation.

        Parameters:
        name (str): Name of the metric.
        metric_class (class): Class of the metric.
        """
        self.metrics[name] = metric_class()
        self.results[name] = []

    def update(self, y, Gamma=None, p_values=None, cpd=None, raise_errors=True):
        """
        Update the metrics with new data.

        Parameters:
        y: True label.
        Gamma: Prediction set.
        p_values: Probability values.
        cpd: Conformal predictive distribution.
        epsilon: Significance level.
        raise_errors (bool): Whether to raise errors for missing data.
        """
        for name, metric in self.metrics.items():
            try:
                if isinstance(metric, OE) or isinstance(metric, Err) or isinstance(metric, Width) or isinstance(metric, WinklerScore):
                    if Gamma is None:
                        if raise_errors:
                            raise ValueError(f"Gamma (gamma_val) is required for {name}.")
                        else:
                            continue

                    if isinstance(metric, WinklerScore):
                        result = metric._update(Gamma, y, Gamma.epsilon)  # Correct arguments for WinklerScore
                    elif isinstance(metric, Width):
                        result = metric._update(Gamma) # Correct arguments for Width
                    else:
                        result = metric._update(y, Gamma)      # Correct arguments for OE, Err
                elif isinstance(metric, OF):
                    if p_values is None:
                        if raise_errors:
                            raise ValueError(f"Probability values (p_vals) are required for {name}.")
                        else:
                            continue
                    result = metric._update(p_values, y)           # Correct arguments for OF
                elif isinstance(metric, CRPS):
                    if cpd is None:
                        if raise_errors:
                            raise ValueError(f"CPD (cpd) is required for {name}.")
                        else:
                            continue
                    result = metric._update(cpd, y)              # Correct arguments for CRPS
                else:
                    raise ValueError(f"Unknown metric type: {type(metric)}")

                self.results[name].append(result)  # Store the result

            except ValueError as e:
                if raise_errors:
                    raise
                else:
                    print(f"Warning: Skipping metric '{name}': {e}")
        self.n += 1
        
    def summarize(self):
        """
        Summarize the results of the metrics.

        Returns:
        dict: Summary of the metrics.
        """
        summary = {}
        for name, results in self.results.items():
            summary[name] = {
                "mean": np.mean(results) if results else None,
                "std": np.std(results) if results else None,
                "min": np.min(results) if results else None,
                "max": np.max(results) if results else None,
                "last": results[-1] if results else None,
            }
        return summary
    
    def cumulative(self, criterion):
        """
        Calculate the cumulative sum of a criterion.

        Parameters:
        criterion (str): Name of the criterion.

        Returns:
        np.ndarray: Cumulative sum of the criterion.
        """
        return np.cumsum(self.results[criterion])
    
    def __getattr__(self, name):  # Called when an attribute is not found
        """
        Get a metric by name.

        Parameters:
        name (str): Name of the metric.

        Returns:
        Metric instance.
        """
        if name in self.metrics:
            return self.metrics[name]
        else:
            raise AttributeError(f"Metric '{name}' not found in results.")

    def __dir__(self): #For autocompletion
        """
        List all available metrics.

        Returns:
        list: List of metric names.
        """
        return list(self.metrics.keys())
    
    
    # FIXME: I want to have x tics on the x-axis, but the same x-axis for all plots
    def plot_cumulative_results(self, max_cols=2, title=None):
        """
        Plot the cumulative results of the metrics.

        Parameters:
        max_cols (int): Maximum number of columns in the plot.
        title (str): Title of the plot.

        Returns:
        matplotlib.figure.Figure: Figure object of the plot.
        """
        num_criteria = len(self.results)
        
        # Determine grid size
        num_cols = min(max_cols, num_criteria)  # Limit columns
        num_rows = int(np.ceil(num_criteria / num_cols))

        # Adaptive figure size
        fig, axs = plt.subplots(num_rows, num_cols, figsize=(5 * num_cols, 3 * num_rows), squeeze=False, sharex=False)

        # Flatten axs for easy indexing if it's multidimensional
        axs = axs.flatten()

        for i, (criterion, values) in enumerate(self.results.items()):
            cs = np.where(np.isinf(values), np.nan, np.cumsum(values))
            axs[i].plot(cs)
            # axs[i].set_title(criterion)
            axs[i].set_xlabel('Number of examples')
            axs[i].set_ylabel(criterion)

        # Hide any unused subplots
        for j in range(i + 1, len(axs)):
            fig.delaxes(axs[j])

        if title is not None:
            fig.suptitle(title)
        fig.tight_layout()
        plt.close(fig)  # Prevent implicit display
        return fig


class EfficiencyCriterion:
    """
    Base class for efficiency criteria.
    """

    def __init__(self):
        """
        Initialize the EfficiencyCriterion class.
        """
        self.value = 0.0
        self.n = 0

    def mean(self):
        """
        Calculate the mean value of the criterion.

        Returns:
        float: Mean value.
        """
        return self.value / self.n

class OE(EfficiencyCriterion):
    """
    Observed excess (OE) is the number of incorrect labels included in the prediction set Gamma.
    Thus, it depends on the significance level varepsilon.
    OE is a conditionally proper efficiency criterion (see Algorithmic Learning in a Random World 2nd edition).
    """
    def __init__(self):
        """
        Initialize the OE class.
        """
        super().__init__()

    def _update(self, y, Gamma):
        """
        Update the OE metric with new data.

        Parameters:
        y: True label.
        Gamma: Prediction set.

        Returns:
        int: Updated OE value.
        """
        self.n += 1
        if y in Gamma:
            oe = len(Gamma) - 1
        else:
            oe = len(Gamma)
        self.value += oe
        return oe
    
class OF(EfficiencyCriterion):
    """
    Observed fuzzyness (OF) is the sum of incorrect p-values. Thus, it is independent of the significance level varepsilon.
    OF is a conditionally proper efficiency criterion (see Algorithmic Learning in a Random World 2nd edition).
    """
    def __init__(self):
        """
        Initialize the OF class.
        """
        super().__init__()

    def _update(self, p_values, y):
        """
        Update the OF metric with new data.

        Parameters:
        p_values (dict): p-values.
        y: True label.

        Returns:
        float: OF value.
        """
        self.n += 1
        of = 0
        for label, p in p_values.items():
            if not label == y:
                of += p
        self.value += of
        return of
    
class Err(EfficiencyCriterion):
    """
    Err evaluates whether the true label is included in the prediction set Gamma.
    """

    def __init__(self):
        """
        Initialize the Err class.
        """
        super().__init__()

    def _update(self, y, Gamma):
        """
        Update the Err metric with new data.

        Parameters:
        y: True label.
        Gamma: Prediction set.

        Returns:
        int: Updated Err value.
        """
        self.n += 1
        if not type(Gamma) == tuple:
            err = int(not(y in Gamma))
        else:       
            err = int(not(y in Gamma))
        self.value += err
        return err

    
class Width(EfficiencyCriterion):
    """
    The width of the prediciton interval Gamma.
    """
    def __init__(self):
        """
        Initialize the Width class.
        """
        super().__init__()
    
    def _update(self, Gamma):
        """
        Update the Width metric with new data.

        Parameters:
        Gamma: Prediction set.

        Returns:
        float: Width value.
        """
        self.n += 1
        w = Gamma.width()
        self.value += w
        return w
    
class WinklerScore(EfficiencyCriterion):
    """
    The Winkler interval score (or just interval score) is a proper scoring rule for evaluation interval forecasts.
    """

    def __init__(self):
        """
        Initialize the WinklerScore class.
        """
        super().__init__()

    def _update(self, Gamma, y, epsilon=0.1):
        """
        Update the WinklerScore metric with new data.

        Parameters:
        Gamma: Prediction set.
        y: True label.
        epsilon (float): Significance level.

        Returns:
        float: WinklerScore value.
        """
        epsilon = Gamma.epsilon
        self.n += 1
        try:
            assert Gamma.upper < np.inf
            assert Gamma.lower > -np.inf
        except AssertionError:
            ws = np.inf
            self.value += ws
            return ws
        l = Gamma.lower
        u = Gamma.upper
        base_score = u - l
        penalty = 0
        
        if y < l:
            penalty = (2 / epsilon) * (l - y)
        elif y > u:
            penalty = (2 / epsilon) * (y - u)
        
        ws = base_score + penalty

        self.value += ws
        return ws
    
class CRPS(EfficiencyCriterion):
    """
    Continuous ranked probability score (CRPS) is a proper scoring rule for evaluating predictive distributions.
    In the case of conformal predictive distributions, the definition does not work, as the integral will not 
    converge. To address this, we have pieced together a confergent integral by integrating the lower distribution
    when x <= y, and the upper when x > y.
    """

    def __init__(self):
        """
        Initialize the CRPS class.
        """
        super().__init__()

    def _update(self, cpd, y):
        """
        Update the CRPS metric with new data.

        Parameters:
        cpd: Conformal predictive distribution.
        y: True label.

        Returns:
        float: CRPS value.
        """
        '''
        NOTE: This implementation is a bith shaky, really. I integrate the lower distribution for x <= y, 
              and the upper for x > y, to ensure convergence. It is possible that this is not proper.
        '''
        self.n += 1
        func_ad_hoc = lambda x: (cpd(x, 0) - int(y <= x))**2 if x <= y else (cpd(x, 1) - int(y <= x))**2

        crps = np.trapz([func_ad_hoc(x) for x in cpd.y_vals[1:-1]], cpd.y_vals[1:-1]) # NOTE: Perhaps not the best integrator

        self.value += crps
        return crps
    
if __name__ == "__main__":
    import doctest
    import sys
    (failures, _) = doctest.testmod()
    if failures:
        sys.exit(1)



import numpy as np
from scipy.integrate import quad
from scipy.optimize import minimize
from scipy.special import betaln
from scipy.stats import beta, norm
import warnings


class BettingStrategy:
    """
    Base class for betting strategies.

    Attributes:
        min_sample_size (int): Minimum number of samples required before full betting.
        mixing_exponent (float): Exponent controlling the mixing parameter's growth.
        mixing_parameter (float): Mixing parameter for cautious betting.
    """

    def __init__(self, min_sample_size=100, mixing_exponent=1):
        """
        Initializes the BettingStrategy class.

        Args:
            min_sample_size (int): Minimum number of samples required before full betting.
            mixing_exponent (float): Exponent controlling the mixing parameter's growth.
        """
        self.min_sample_size = min_sample_size
        self.mixing_exponent = mixing_exponent
        self.mixing_parameter = 0
    
    def update_mixing_parameter(self, n):
        """
        Updates the mixing parameter based on the number of observations.

        Args:
            n (int): Number of observations.
        """
        self.mixing_parameter = min((n / self.min_sample_size)**self.mixing_exponent, 1)


class GaussianKDE(BettingStrategy):
    """
    Implements a Gaussian Kernel Density Estimation (KDE) betting strategy.

    Attributes:
        bandwidth (str or float): Bandwidth selection method or fixed value.
        window_size (str or int): Window size for adaptive KDE.
    """

    def __init__(self, bandwidth='silverman', window_size=None, min_sample_size=100, mixing_exponent=1):
        """
        Initializes the GaussianKDE class.

        Args:
            bandwidth (str or float): Bandwidth selection method or fixed value.
            window_size (str or int): Window size for adaptive KDE.
            min_sample_size (int): Minimum number of samples required before full betting.
            mixing_exponent (float): Exponent controlling the mixing parameter's growth.
        """
        super().__init__(min_sample_size, mixing_exponent)
        self.bandwidth = bandwidth
        self.window_size = window_size

    def calculate_bandwidth(self, data, sigma=None):
        """
        Calculates the bandwidth for KDE.

        Args:
            data (np.ndarray): Data points.
            sigma (float): Standard deviation of the data.

        Returns:
            float: Bandwidth value.
        """
        if self.bandwidth == 'silverman':
            assert sigma is not None
            h = ((4 * sigma**5) / (3 * data.size))**(1/5)
        else:
            h = self.bandwidth
        return h
    
    def calculate_window_size(self, p_values):
        """
        Calculates the window size for adaptive KDE.

        Args:
            p_values (list): List of p-values.

        Returns:
            int: Window size.
        """
        if self.window_size == 'adaptive':
            window_param = np.log(1.001)
            min_size = self.min_sample_size
            max_size = len(p_values)
            return max(min_size, int(max_size * np.exp(-window_param * self.M)))
        elif self.window_size is None:
            return 0
        else:
            return self.window_size

    def update_betting_function(self, p_values):
        """
        Updates the betting function based on the provided p-values.

        Args:
            p_values (list): List of p-values.

        Returns:
            tuple: Updated betting functions (b_n, B_n).
        """
        if len(p_values) < 2:  # Too little information to bet
            pdf = lambda x: beta.pdf(x, 1, 1)
            cdf = lambda x: beta.cdf(x, 1, 1)
        else:
            data = np.array(p_values)[-self.calculate_window_size(p_values):]
            # sigma = np.array([data, 2 - data, -data]).flatten().std()
            sigma = np.array(data).std()
            if sigma == 0:
                # If there is no variability: do not bet at all.
                pdf = lambda x: beta.pdf(x, 1, 1)
                cdf = lambda x: beta.cdf(x, 1, 1)
            else:
                h = self.calculate_bandwidth(data=data, sigma=sigma)

                def kernel_pdf_raw(x):
                    """
                    Computes the raw kernel PDF.

                    Args:
                        x (np.ndarray): Input values.

                    Returns:
                        np.ndarray or float: PDF values.
                    """
                    x = np.atleast_1d(x)
                    pdf_values = np.mean(norm.pdf(x[:, None], loc=data, scale=h * sigma), axis=1)
                    if pdf_values.size == 1:
                        return pdf_values.item()
                    return pdf_values

                def kernel_cdf_raw(x):
                    """
                    Computes the raw kernel CDF.

                    Args:
                        x (np.ndarray): Input values.

                    Returns:
                        np.ndarray or float: CDF values.
                    """
                    x = np.atleast_1d(x)
                    cdf_values = np.mean(norm.cdf(x[:, None], loc=data, scale=h * sigma), axis=1)
                    if cdf_values.size == 1:
                        return cdf_values.item()
                    return cdf_values

                def kernel_pdf_reflect(x):
                    """
                    Computes the reflected kernel PDF.

                    Args:
                        x (np.ndarray): Input values.

                    Returns:
                        np.ndarray or float: Reflected PDF values.
                    """
                    x = np.atleast_1d(x)
                    pdf_reflect_values = kernel_pdf_raw(x) + kernel_pdf_raw(-x) + kernel_pdf_raw(2 - x)
                    if np.isscalar(pdf_reflect_values):
                        return pdf_reflect_values
                    return pdf_reflect_values

                def kernel_cdf_reflect(x):
                    """
                    Computes the reflected kernel CDF.

                    Args:
                        x (np.ndarray): Input values.

                    Returns:
                        np.ndarray or float: Reflected CDF values.
                    """
                    x = np.atleast_1d(x)
                    cdf_reflect_values = kernel_cdf_raw(x) - kernel_cdf_raw(-x) + 1 - kernel_cdf_raw(2 - x)
                    if np.isscalar(cdf_reflect_values):
                        return cdf_reflect_values
                    return cdf_reflect_values
                
                pdf = kernel_pdf_reflect
                cdf = kernel_cdf_reflect
        
        # Cautious betting until min_sample_size is reached
        b_n = lambda x: self.mixing_parameter * pdf(x) + (1 - self.mixing_parameter)
        B_n = lambda x: self.mixing_parameter * cdf(x) + (1 - self.mixing_parameter) * x
        
        self.update_mixing_parameter(len(p_values))

        return b_n, B_n


class BetaMoments(BettingStrategy):
    """
    Implements a betting strategy based on Beta distribution moments.

    Attributes:
        n (int): Number of observations.
        mean (float): Running mean of observations.
        M2 (float): Sum of squared differences from the mean (for variance).
        ahat (float): Alpha parameter of the Beta distribution.
        bhat (float): Beta parameter of the Beta distribution.
    """

    def __init__(self, min_sample_size=100, mixing_exponent=1):
        """
        Initializes the BetaMoments class.

        Args:
            min_sample_size (int): Minimum number of samples required before full betting.
            mixing_exponent (float): Exponent controlling the mixing parameter's growth.
        """
        super().__init__(min_sample_size, mixing_exponent)
        self.n = 0
        self.mean = 0.0
        self.M2 = 0.0
        self.ahat = 1.0
        self.bhat = 1.0

    def update_betting_function(self, p):
        """
        Updates the betting function based on the provided p-value.

        Args:
            p (float): A single p-value.

        Returns:
            tuple: Updated betting functions (b_n, B_n).
        """
        if self.n < 2:
            self.ahat = 1.0
            self.bhat = 1.0
        else:
            sample_variance = self.M2 / (self.n - 1) if self.n > 1 else 0
            if sample_variance <= 0:
                self.ahat = 1
                self.bhat = 1
            else:
                common_factor = (self.mean * (1 - self.mean) / sample_variance) - 1
                self.ahat = self.mean * common_factor
                self.bhat = (1 - self.mean) * common_factor

        self.n += 1
        delta = p - self.mean
        self.mean += delta / self.n
        delta2 = p - self.mean
        self.M2 += delta * delta2

        b_n = lambda x: self.mixing_parameter * beta.pdf(x, self.ahat, self.bhat) + (1 - self.mixing_parameter)
        B_n = lambda x: self.mixing_parameter * beta.cdf(x, self.ahat, self.bhat) + (1 - self.mixing_parameter) * x
        
        self.update_mixing_parameter(self.n)

        return b_n, B_n


class BetaMLE(BettingStrategy):
    """
    Implements a betting strategy based on Maximum Likelihood Estimation (MLE) for Beta distribution parameters.

    Attributes:
        n (int): Number of observations.
        log_sum_x (float): Sum of logarithms of observations.
        log_sum_1_minus_x (float): Sum of logarithms of (1 - observations).
        ahat (float): Alpha parameter of the Beta distribution.
        bhat (float): Beta parameter of the Beta distribution.
    """

    def __init__(self, min_sample_size=100, mixing_exponent=1):
        """
        Initializes the BetaMLE class.

        Args:
            min_sample_size (int): Minimum number of samples required before full betting.
            mixing_exponent (float): Exponent controlling the mixing parameter's growth.
        """
        super().__init__(min_sample_size, mixing_exponent)
        self.n = 0
        self.log_sum_x = 0.0
        self.log_sum_1_minus_x = 0.0
        self.ahat = 1.0
        self.bhat = 1.0

    def update_betting_function(self, p):
        """
        Updates the betting function based on the provided p-value.

        Args:
            p (float): A single p-value.

        Returns:
            tuple: Updated betting functions (b_n, B_n).
        """
        def negative_log_likelihood(params):
            alpha, beta = params
            if alpha <= 0 or beta <= 0:
                return np.inf
            log_likelihood = (
                (alpha - 1) * self.log_sum_x +
                (beta - 1) * self.log_sum_1_minus_x -
                self.n * betaln(alpha, beta)
            )
            return -log_likelihood

        if self.n < 2:
            self.ahat = 1.0
            self.bhat = 1.0
        else:
            initial_guess = [1.0, 1.0]
            result = minimize(negative_log_likelihood, initial_guess, bounds=[(1e-5, None), (1e-5, None)])
            if result.success:
                self.ahat, self.bhat = result.x
            else:
                self.ahat, self.bhat = 1.0, 1.0

        self.n += 1
        self.log_sum_x += np.log(p)
        self.log_sum_1_minus_x += np.log(1 - p)

        b_n = lambda x: self.mixing_parameter * beta.pdf(x, self.ahat, self.bhat) + (1 - self.mixing_parameter)
        B_n = lambda x: self.mixing_parameter * beta.cdf(x, self.ahat, self.bhat) + (1 - self.mixing_parameter) * x
        
        self.update_mixing_parameter(self.n)

        return b_n, B_n


class ConformalTestMartingale:
    """
    Parent class for conformal test martingales.

    Attributes:
        logM (float): Logarithm of the martingale value.
        max (float): Maximum martingale value observed so far.
        p_values (list): List of observed p-values.
        log_martingale_values (list): Logarithm of martingale values over time.
        warning_level (float): Threshold for raising warnings about exchangeability violations.
        warnings (bool): Whether to raise warnings when the threshold is exceeded.
        b_n (function): Current betting function for density.
        B_n (function): Current betting function for cumulative density.
    """

    def __init__(self, warnings=True, warning_level=100):
        """
        Initializes the ConfromalTestMartingale class.

        Args:
            warnings (bool): Whether to raise warnings when the threshold is exceeded.
            warning_level (float): Threshold for raising warnings about exchangeability violations.
        """
        self.logM = 0.0
        self.max = 1.0
        self.p_values = []
        self.log_martingale_values = [0.0]
        self.warning_level = warning_level
        self.warnings = warnings
        self.b_n = lambda x: beta.pdf(x, 1, 1)
        self.B_n = lambda x: beta.cdf(x, 1, 1)

    @property
    def M(self):
        """
        Returns the current martingale value.

        Returns:
            float: Martingale value.
        """
        return np.exp(self.logM)
    
    @property
    def martingale_values(self):
        """
        Returns the martingale values over time.

        Returns:
            list: Martingale values.
        """
        return np.exp(self.log_martingale_values)
    
    @property
    def log10_martingale_values(self):
        """
        Returns the base-10 logarithm of martingale values over time.

        Returns:
            list: Log10 martingale values.
        """
        return np.log10(self.martingale_values)
    
    def check_warning(self):
        """
        Checks if the martingale value exceeds the warning threshold and raises a warning if necessary.
        """
        if self.max >= self.warning_level and self.warnings:
            warnings.warn(f'Exchangeability assumption likely violated: Max martingale value is {self.max}')


class PluginMartingale(ConformalTestMartingale):
    """
    Implements a plugin martingale using a specified betting strategy.

    Attributes:
        betting_strategy (BettingStrategy): The betting strategy to use.
    """

    def __init__(self, betting_strategy=GaussianKDE, warnings=True, warning_level=100, **kwargs):
        """
        Initializes the PluginMartingale class.

        Args:
            betting_strategy (BettingStrategy or type): The betting strategy to use.
            warnings (bool): Whether to raise warnings when the threshold is exceeded.
            warning_level (float): Threshold for raising warnings about exchangeability violations.
            **kwargs: Additional arguments for the betting strategy.
        """
        super().__init__(warnings, warning_level)

        if isinstance(betting_strategy, BettingStrategy):
            self.betting_strategy = betting_strategy
        else:
            betting_kwargs = kwargs if kwargs else {
                'bandwidth': 'silverman', 
                'window_size': None, 
                'min_sample_size': 100, 
                'mixing_exponent': 1.
            }
            self.betting_strategy = betting_strategy(**betting_kwargs)

    def update_martingale_value(self, p):
        """
        Updates the martingale value based on the provided p-value.

        Args:
            p (float): A single p-value.
        """
        self.logM += np.log(self.b_n(p))
        self.log_martingale_values.append(self.logM)
        self.p_values.append(p)
        
        if isinstance(self.betting_strategy, GaussianKDE):
            self.b_n, self.B_n = self.betting_strategy.update_betting_function(self.p_values)
        else:
            self.b_n, self.B_n = self.betting_strategy.update_betting_function(p)
        
        if self.M > self.max:
            self.max = self.M

        self.check_warning()


class SimpleJumper(ConformalTestMartingale):
    """
    Implements a simple jumper martingale.

    Attributes:
        J (float): Jump size parameter.
        C_epsilon (dict): Dictionary of martingale values for different epsilon values.
        C (float): Combined martingale value.
        b_epsilon (function): Betting function for density.
        B_n_inv (function): Inverse of the cumulative betting function.
    """

    def __init__(self, J=0.01, warning_level=100, warnings=True, **kwargs):
        """
        Initializes the SimpleJumper class.

        Args:
            J (float): Jump size parameter.
            warning_level (float): Threshold for raising warnings about exchangeability violations.
            warnings (bool): Whether to raise warnings when the threshold is exceeded.
            **kwargs: Additional arguments.
        """
        super().__init__(warnings, warning_level)
        self.J = J
        self.C_epsilon = {-1: 1/3, 0: 1/3, 1: 1/3}
        self.C = 1
        self.b_epsilon = lambda u, epsilon: 1 + epsilon * (u - 1/2)
        self.B_n_inv = lambda x: x
    
    def update_martingale_value(self, p):
        """
        Updates the martingale value based on the provided p-value.

        Args:
            p (float): A single p-value.
        """
        self.p_values.append(p)
        for epsilon in [-1, 0, 1]:
            self.C_epsilon[epsilon] = (1 - self.J) * self.C_epsilon[epsilon] + (self.J / 3) * self.C
            self.C_epsilon[epsilon] = self.C_epsilon[epsilon] * self.b_epsilon(p, epsilon)
        self.C = self.C_epsilon[-1] + self.C_epsilon[0] + self.C_epsilon[1]
        self.logM = np.log(self.C)
        self.log_martingale_values.append(self.logM)

        epsilon_bar = (self.C_epsilon[1] - self.C_epsilon[-1]) / self.C
        self.b_n = lambda u: 1 + epsilon_bar * (u - 1/2)
        self.B_n = lambda u: (epsilon_bar / 2) * u**2 + (1 - epsilon_bar / 2) * u
        self.B_n_inv = lambda u: (epsilon_bar - 2) / (2 * epsilon_bar) + np.sqrt(epsilon_bar * (8 * u + epsilon_bar - 4) + 4) / (2 * epsilon_bar)

        if self.M > self.max:
            self.max = self.M

        self.check_warning()

        
if __name__ == "__main__":
    import doctest
    import sys
    (failures, _) = doctest.testmod()
    if failures:
        sys.exit(1)



import numpy as np
import time
import warnings
from scipy.spatial.distance import pdist, cdist, squareform
import matplotlib.pyplot as plt
from scipy.optimize import minimize_scalar, minimize, Bounds
MACHINE_EPSILON = lambda x: np.abs(x) * np.finfo(np.float64).eps


default_epsilon = 0.1


def get_ConformalPredictionInterval():
    from .regressors import ConformalPredictionInterval  # Lazy import
    return ConformalPredictionInterval

class ConformalPredictiveSystem:
    '''
    Parent class for conformal predictive systems. Unclear if some methods are common to all, so perhaps we don't need it.
    '''

    def __init__(self, epsilon=default_epsilon):
        self.epsilon = epsilon

    def learn_many(self, X, y):
        for x1, y1 in zip(X, y):
            self.learn_one(x1,y1)

class RidgePredictionMachine(ConformalPredictiveSystem):
    'This conformal predictive system uses the "studentised residuals as conformity measure'

    def __init__(self, a=0, warnings=True, autotune=False, verbose=0, rnd_state=None, epsilon=default_epsilon):
        super().__init__(epsilon=epsilon)
        self.a = a
        self.X = None
        self.y = None
        self.p = None
        self.Id = None
        self.XTXinv = None

        # Should we raise warnings
        self.warnings = warnings
        # Do we autotune ridge prarmeter on warning
        self.autotune = autotune

        self.verbose = verbose
        self.rnd_gen = np.random.default_rng(rnd_state)

    def learn_initial_training_set(self, X, y):
        self.X = X
        self.y = y
        self.p = X.shape[1]
        self.Id = np.identity(self.p)
        if self.autotune:
            self._tune_ridge_parameter()
        else:
            self.XTXinv = np.linalg.inv(self.X.T @ self.X + self.a*self.Id)

    def learn_one(self, x, y, precomputed=None):
        '''
        Learn a single example. If we have already computed X and XTXinv, use them for update. Then the last row of X is the object with label y.
        >>> cps = RidgePredictionMachine()
        >>> cps.learn_one(np.array([1,0]), 1)
        >>> cps.X
        array([[1, 0]])
        >>> cps.y
        array([1])
        '''
        # Learn label y
        if self.y is None:
            self.y = np.array([y])
        else:
            if hasattr(self, 'h'):
                self.y = np.append(self.y, y.reshape(1, self.h), axis=0)
            else:
                self.y = np.append(self.y, y)
    
        if precomputed is not None:
            X = precomputed['X']
            XTXinv = precomputed['XTXinv']

            if X is not None:
                self.X = X
                self.p = self.X.shape[1]
                self.Id = np.identity(self.p)
            
            if XTXinv is not None:
                self.XTXinv = XTXinv
                
            else:
                if self.X.shape[0] == 1:
                    # print(self.X)
                    # print(self.Id)
                    # print(self.a)
                    self.XTXinv = np.linalg.inv(self.X.T @ self.X + self.a * self.Id)
                else:
                    # Update XTX_inv (inverse of Kernel matrix plus regularisation) Use the Sherman-Morrison formula to update the hat matrix
                            #https://en.wikipedia.org/wiki/Sherman%E2%80%93Morrison_formula
                    self.XTXinv -= (self.XTXinv @ np.outer(x, x) @ self.XTXinv) / (1 + x.T @ self.XTXinv @ x)
            
                    # Check the rank
                    if self.warnings:
                        rank_deficient = not(self.check_matrix_rank(self.XTXinv))
                    
        else:
            # Learn object x
            if self.X is None:
                self.X = x.reshape(1,-1)
                self.p = self.X.shape[1]
                self.Id = np.identity(self.p)
            elif self.X.shape[0] == 1:
                self.X = np.append(self.X, x.reshape(1, -1), axis=0)
                self.XTXinv = np.linalg.inv(self.X.T @ self.X + self.a * self.Id)
            else:
                self.X = np.append(self.X, x.reshape(1, -1), axis=0)
                # Update XTX_inv (inverse of Kernel matrix plus regularisation) Use the Sherman-Morrison formula to update the hat matrix
                        #https://en.wikipedia.org/wiki/Sherman%E2%80%93Morrison_formula
                self.XTXinv -= (self.XTXinv @ np.outer(x, x) @ self.XTXinv) / (1 + x.T @ self.XTXinv @ x)
        
                # Check the rank
                if self.warnings:
                    rank_deficient = not(self.check_matrix_rank(self.XTXinv))


    def change_ridge_parameter(self, a):
        '''
        Change the ridge parameter
        >>> cps = RidgePredictionMachine()
        >>> cps.learn_one(np.array([1,0]), 1)
        >>> cps.change_ridge_parameter(1)
        >>> cps.a
        1
        '''
        self.a = a
        if self.X is not None:
            self.XTXinv = np.linalg.inv(self.X.T @ self.X + self.a * self.Id)


    def check_matrix_rank(self, M):
        '''
        Check if a matrix has full rank <==> is invertible
        Returns False if matrix is rank deficient
        NOTE In numerical linear algebra it is a bit more subtle. The condition number can tell us more.

        >>> cps = RidgePredictionMachine(warnings=False)
        >>> cps.check_matrix_rank(np.array([[1, 0], [1, 0]]))
        False
        >>> cps.check_matrix_rank(np.array([[1, 0], [0, 1]]))
        True
        '''
        if np.linalg.matrix_rank(M) < M.shape[0]:
            if self.warnings:
                warnings.warn(f'The matrix X is rank deficient. Condition number: {np.linalg.cond(M)}. Consider changing the ridge prarmeter')
            return False
        else:
            return True


    def _tune_ridge_parameter(self, a0=None):
        '''
        Tune ridge parameter with Generalized cross validation https://pages.stat.wisc.edu/~wahba/stat860public/pdf1/golub.heath.wahba.pdf
        '''
        XTX = self.X.T @ self.X
        n = self.X.shape[0]
        In = np.identity(n)

        def GCV(a):
            try:
                A = self.X @ np.linalg.inv(XTX + a*self.Id) @ self.X.T
                max_diag_H = np.max(np.diag(A))  # Maximum diagonal element of the hat matrix
                if max_diag_H > 1:
                    return np.inf 
                return (1/n)*np.linalg.norm((In - A) @ self.y)**2 / ((1/n)* np.trace(In- A))**2
            except (np.linalg.LinAlgError, ZeroDivisionError):
                return np.inf
        
        # Initial guess
        if a0 is None:
            a0 = 1e-6 # Just a small pertubation to avoid numerical issues

        # Bounds to ensure a >= 0
        res = minimize(GCV, x0=a0, bounds=Bounds(lb=1e-6, keep_feasible=True)) # May be relevant to pass some arguments here, or even use another minimizer.
        a = res.x[0]

        if self.verbose > 0:
            print(f'New ridge parameter: {a}')
        self.change_ridge_parameter(a)


    def predict_cpd(self, x, return_update=False, save_time=False):
        def build_precomputed(X, XTXinv, C):
            computed = {
                'X': X, # The updated matrix of objects
                'XTXinv': XTXinv, # The updated kernel matrix
                'C': C
            } 
            return computed

        tic = time.time()
        # Add row to X matrix
        X = np.append(self.X, x.reshape(1, -1), axis=0)
        toc_add_row = time.time() - tic
        n = X.shape[0]
        y = self.y

        tic = time.time()
        # Update XTX_inv (inverse of Kernel matrix plus regularisation) Use the Sherman-Morrison formula to update the hat matrix
                #https://en.wikipedia.org/wiki/Sherman%E2%80%93Morrison_formula

        XTXinv = self.XTXinv - (self.XTXinv @ np.outer(x, x) @ self.XTXinv) / (1 + x.T @ self.XTXinv @ x)
        toc_update_XTXinv = time.time() - tic

        tic = time.time()

        H = X @ XTXinv @ X.T
        h = H.diagonal()
        sqrt_one_minus_h = np.sqrt(1 - h[:-1])
        A = np.dot(H[-1, :-1], y) / np.sqrt(1 - h[-1])  + (y - H[:-1, :-1] @ y) / sqrt_one_minus_h
        B = np.sqrt(1 - h[-1]) * np.ones(n-1) + H[-1, :-1] / sqrt_one_minus_h
        C = np.zeros(n + 1)
        C[1:-1] = A / B
        C[0] = -np.inf
        C[-1] = np.inf
        C.sort()


        # NOTE: Keep the loop version for a bit, if we run into problems...
        # C = np.empty(n+1)
        # for i in range(n - 1):
        #     a_i = sum([H[j, -1] * y[j] for j in range(n-1)]) / np.sqrt(1 - h[-1]) + (y[i] - sum([H[i, j] * y[j] for j in range(n-1)])) / np.sqrt(1 - h[i])
        #     b_i = np.sqrt(1 - h[-1]) + H[i, -1] / np.sqrt(1 - h[i])
        #     C[i+1] = a_i / b_i
        # C[0] = -np.inf
        # C[-1] = np.inf
        # C.sort()

        toc_compute_C = time.time() - tic

        time_dict = {
            'Update hat matrix': toc_update_XTXinv,
            'Compute C': toc_compute_C
        }
        time_dict = time_dict if save_time else None
        cpd = RidgePredictiveDistributionFunction(C=C, time_dict=time_dict, epsilon=self.epsilon)

        if return_update:
            return cpd, build_precomputed(X, XTXinv, C)
        else:
            return cpd
    

class KernelRidgePredictionMachine(ConformalPredictiveSystem):
    '''
        This conformal predictive system uses the "studentised residuals as conformity measure.
        Algorithm 7,3 in Algorithmic Learning in a Random World 2nd edition.
    '''

    def __init__(self, kernel, a=0, warnings=True, autotune=False, verbose=0, rnd_state=None, epsilon=default_epsilon):
        super().__init__(epsilon=epsilon)

        self.kernel = kernel

        self.a = a
        self.X = None
        self.y = None
        

        # Should we raise warnings
        self.warnings = warnings
        # Do we autotune ridge prarmeter on warning
        self.autotune = autotune

        self.verbose = verbose
        self.rnd_gen = np.random.default_rng(rnd_state)

    def learn_initial_training_set(self, X, y):
        self.X = X
        self.y = y
        Id = np.identity(self.X.shape[0])
        self.K = self.kernel(self.X)
        if self.autotune:
            self._tune_ridge_parameter()
        else:
            self.Kinv = np.linalg.inv(self.K + self.a * Id)
        self.H = self.K @ self.Kinv

    def _tune_ridge_parameter(self, a0=None):
        '''
        Tune ridge parameter with Generalized Cross Validation (GCV) in the kernel space.
        '''
        n = self.K.shape[0]
        In = np.identity(n)

        def GCV(a):
            try:
                A = self.K @ np.linalg.inv(self.K + a * In)
                max_diag_H = np.max(np.diag(A))  # Maximum diagonal element of the hat matrix
                if max_diag_H > 1:
                    return np.inf 
                return (1 / n) * np.linalg.norm((In - A) @ self.y) ** 2 / ((1 / n) * np.trace(In - A)) ** 2
            except (np.linalg.LinAlgError, ZeroDivisionError):
                return np.inf

        # Initial guess
        if a0 is None:
            a0 = 1e-6  # Small perturbation to avoid numerical issues

        # Bounds to ensure a >= 0
        res = minimize(GCV, x0=a0, bounds=Bounds(lb=1e-6, keep_feasible=True))
        a = res.x[0]

        if self.verbose > 0:
            print(f'New ridge parameter: {a}')
        self.change_ridge_parameter(a)

    def change_ridge_parameter(self, a):
        '''
        Change the ridge parameter
        >>> cps = RidgePredictionMachine()
        >>> cps.learn_one(np.array([1,0]), 1)
        >>> cps.change_ridge_parameter(1)
        >>> cps.a
        1
        '''
        self.a = a
        if self.X is not None:
            Id = np.identity(self.X.shape[0])
       
            self.K = self.kernel(self.X)
            self.Kinv = np.linalg.inv(self.K + self.a * Id)
            self.H = self.K @ self.Kinv

    def _update_Kinv(self, Kinv, k, d):
        # print(f'K: {K}')
        # print(f'k: {k}')
        # print(f'kappa: {kappa}')
        return np.block([[Kinv + d * Kinv @ k @ k.T @ Kinv, -d * Kinv @ k], [ -d * k.T @ Kinv, d]])


    @staticmethod
    def _update_K(K, k, kappa):
        # print(f'K: {K}')
        # print(f'k: {d}')
        # print(f'kappa: {kappa}')
        return np.block([[K, k], [k.T, kappa]])


    def learn_one(self, x, y, precomputed=None):
        '''
        Learn a single example
        '''

        # # TEST DUMB UPDATE
        # self.y = np.append(self.y, y)
        # self.X = np.append(self.X, x.reshape(1, -1), axis=0)
        # Id = np.identity(self.X.shape[0])
       
        # self.K = self.kernel(self.X)
        # self.Kinv = np.linalg.inv(self.K + self.a * Id)
        # self.H = self.K @ self.Kinv

        x = np.atleast_2d(x)
        # Learn label y
        if self.y is None:
            self.y = np.array([y])
        else:
            self.y = np.append(self.y, y)
        
        if precomputed is not None:
            self.H = precomputed['H']
            self.K = self._update_K(self.K, precomputed['k'], precomputed['kappa'])
            self.Kinv = self._update_Kinv(self.Kinv, precomputed['k'], precomputed['d'])
            self.X = np.append(self.X, x.reshape(1, -1), axis=0)
        else:
            if self.X is None:
                self.X = x.reshape(1,-1)
                Id = np.identity(self.X.shape[0])
                self.K = self.kernel(self.X)
                self.Kinv = np.linalg.inv(self.K + self.a * Id)
                self.H = self.K @ self.Kinv
            else:
                k = self.kernel(self.X, x).reshape(-1, 1)
                kappa = self.kernel(x, x)
                d = 1 / (kappa + self.a - k.T @ self.Kinv @ k)
                y = self.y
                self.H = np.block([
                    [
                        self.H - self.a * d * self.Kinv @ k @ k.T @ self.Kinv, 
                        self.a * d * self.Kinv @ k
                    ], 
                    [
                        self.a * d * k.T @ self.Kinv, 
                        d * kappa - d*k.T @ self.Kinv @ k
                        ]
                    ]
                    )
                self.K = self._update_K(self.K, k, kappa)
                self.Kinv = self._update_Kinv(self.Kinv, k, d)
                self.X = np.append(self.X, x.reshape(1, -1), axis=0)
    
    def predict_cpd(self, x, return_update=False, save_time=False):
        
        def build_precomputed(H, k, kappa, d):
            computed = {
                'H': H,
                'k': k,
                'kappa': kappa,
                'd': d,
            } 
            return computed
        x = np.atleast_2d(x)        
        # Temporarily update kernel matrix
        k = self.kernel(self.X, x).reshape(-1, 1)
        kappa = self.kernel(x, x)
        d = 1 / (kappa + self.a - k.T @ self.Kinv @ k)
        y = self.y
        H = np.block([
            [
                self.H - self.a * d * self.Kinv @ k @ k.T @ self.Kinv, 
                self.a * d * self.Kinv @ k
             ], 
             [
                self.a * d * k.T @ self.Kinv, 
                d * kappa - d*k.T @ self.Kinv @ k
                ]
            ]
            )
        
        n = H.shape[0]

        h = H.diagonal()
        sqrt_one_minus_h = np.sqrt(1 - h[:-1])
        A = np.dot(H[-1, :-1], y) / np.sqrt(1 - h[-1])  + (y - H[:-1, :-1] @ y) / sqrt_one_minus_h
        B = np.sqrt(1 - h[-1]) * np.ones(n-1) + H[-1, :-1] / sqrt_one_minus_h

        
        C = np.zeros(n + 1)
        C[1:-1] = A / B
        C[0] = -np.inf
        C[-1] = np.inf
        assert not np.isnan(C).any(), "C contains NaN values"
        C.sort()

        time_dict = None
        cpd = RidgePredictiveDistributionFunction(C=C, time_dict=time_dict, epsilon=self.epsilon)

        if return_update:
            return cpd, build_precomputed(H, k, kappa, d)
        else:
            return cpd

class NearestNeighboursPredictionMachine(ConformalPredictiveSystem):

    def __init__(self, k, distance='euclidean', distance_func=None, warnings=True, verbose=0, rnd_state=None, epsilon=default_epsilon):
        '''
        Consider adding possibility to update self.k as the training set grows, e.g. by some heuristic or something.
        Two rules of thumb are quite simple:
            1. Choose k close to sqrt(n) where n is the training set size
            2. If the data has large variance, choose k larger. If the variance is small, choose k smaller. This is less clear, however.
        '''
        # TODO: The sorting of conformity scores is the most time consuming step. Can it be done with parallel processing to speed things up?
        super().__init__(epsilon=epsilon)

        self.k = k

        self.distance = distance
        if distance_func is None:
            self.distance_func = self._standard_distance_func
        else:
            self.distance_func = distance_func
            self.distance = 'custom'

        self.X = None
        self.y = None
        self.D = None

        # Should we raise warnings
        self.warnings = warnings

        self.verbose = verbose
        self.rnd_gen = np.random.default_rng(rnd_state)


    def _standard_distance_func(self, X, y=None):
        '''
        By default we use scipy to compute distances
        '''
        # TODO: Can this be done using parallel processing? 
        X = np.atleast_2d(X)
        if y is None:
            dists = squareform(pdist(X, metric=self.distance))
        else:
            y = np.atleast_2d(y)
            dists = cdist(X, y, metric=self.distance)
        return dists
    

    def learn_initial_training_set(self, X, y):
        '''
        The Nearest neighbours prediction machine assumes all labels are unique. If they are not, we add noise to break ties.

        >>> cps = NearestNeighboursPredictionMachine(k=3)
        >>> X = np.array([[1], [2]])
        >>> y = np.array([1, 2])
        >>> cps.learn_initial_training_set(X, y)
        >>> cps.X
        array([[1],
               [2]])
        >>> cps.y
        array([1, 2])
        >>> cps.D
        array([[0., 1.],
               [1., 0.]])
        '''
        # FIXME: It also assumes all distances are unique. Figure out how to handle this
        self.X = X
        self.D = self.distance_func(X)

        self.y = y
        
        if np.unique(self.y).size != self.y.size:
            raise Exception('All labels y must be distinct for the NearestNeighboursPredictionMachine to be valid')

    
    @staticmethod
    def update_distance_matrix(D, d):
        return np.block([[D, d], [d.T, np.array([0])]])
    

    def learn_one(self, x, y, precomputed=None):
        '''
        The Nearest neighbours prediction machine assumes all labels are unique. If they are not, we add noise to break ties.
        precomputed is a dictionary
        {
            'X': X,
            'D': D,
        }
        >>> cps = NearestNeighboursPredictionMachine(k=3, rnd_state=2024)
        >>> X = np.array([[1], [2]])
        >>> y = np.array([1, 2])
        >>> cps.learn_initial_training_set(X, y)
        >>> cps.learn_one(np.array([3]), 3)
        >>> cps.y
        array([1, 2, 3])
        >>> cps.X
        array([[1],
               [2],
               [3]])
        >>> cps.D
        array([[0., 1., 2.],
               [1., 0., 1.],
               [2., 1., 0.]])
        '''
        # Learn label y
        if self.y is None:
            self.y = np.array([y])
        else:
            self.y = np.append(self.y, y)

        if precomputed is None:
            # Learn object x
            if self.X is None:
                self.X = x.reshape(1,-1)
                self.D = self.distance_func(self.X)
            else:
                d = self.distance_func(self.X, x)
                self.D = self.update_distance_matrix(self.D, d)
                self.X = np.append(self.X, x.reshape(1, -1), axis=0)
        else:
            self.X = precomputed['X']
            self.D = precomputed['D']

        if np.unique(self.y).size != self.y.size:
            raise Exception('All labels y must be distinct for the NearestNeighboursPredictionMachine to be valid')

    
    def predict_cpd(self, x, return_update=False, save_time=False):
        '''
        >>> import numpy as np
        >>> rnd_gen = np.random.default_rng(2024)
        >>> X = rnd_gen.normal(loc=0, scale=1, size=(100, 4))
        >>> beta = np.array([2, 1, 0, 0])
        >>> Y = X @ beta + rnd_gen.normal(loc=0, scale=1, size=100)
        >>> cps = NearestNeighboursPredictionMachine(k=3)
        >>> cps.learn_initial_training_set(X, Y)
        >>> x = rnd_gen.normal(loc=0, scale=1, size=(1, 4))
        >>> cpd = cps.predict_cpd(x)
        >>> cpd.L
        array([0.        , 0.        , 0.18811881, 0.53465347, 0.76237624])
        >>> cpd.U
        array([0.17821782, 0.18811881, 0.54455446, 0.76237624, 1.        ])
        '''
        tic = time.time()
        # Temporarily update the distance matrix
        if self.X.shape[0] < self.k:
            # FIXME: Make some graceful error handling here
            raise Exception('Training set is too small...')
        else:
            d = self.distance_func(self.X, x)
            D = self.update_distance_matrix(self.D, d)
            y = np.append(self.y, -np.inf) # Initialise label as -inf
        toc_dist = time.time()-tic

        tic = time.time()
        # Find all neighbours and semi-neighbours
        # NOTE: This is the time consuming step. The distance matrix has to be sorted. Is there any way to speed this up?
        k_nearest = D.argsort(axis=0)[1:self.k+1]
        toc_sort = time.time() - tic

        tic = time.time()
        n = self.X.shape[0]

        full_neighbours = []
        single_neighbours = []
        semi_neighbours = []
        idx_all_neighbours_and_semi_neighbours = []

        k_nearest_of_n = k_nearest.T[-1]

        # FIXME: How do we save the full, single and semi-neighbours so that we can acess them later in a nice way?
        for i, col in enumerate(k_nearest.T):
            if i in k_nearest_of_n and n in col:
                # print(f'z_{i} is a full neighbour')
                # idx_all_neighbours_and_semi_neighbours.append(i)
                full_neighbours.append(i)
            if i in k_nearest_of_n and not n in col:
                # print(f'z_{i} is a single neighbour')
                # idx_all_neighbours_and_semi_neighbours.append(i)
                single_neighbours.append(i)
            if not i in k_nearest_of_n and n in col:
                # print(f'z_{i} is a semi-neighbour')
                # idx_all_neighbours_and_semi_neighbours.append(i)
                semi_neighbours.append(i)
        idx_all_neighbours_and_semi_neighbours = np.array(full_neighbours + single_neighbours + semi_neighbours)
        toc_find_neighbours = time.time() - tic
        
        # Line 1
        Kprime = len(idx_all_neighbours_and_semi_neighbours)
        # Line 2 and 3
        Y = np.zeros(shape=Kprime + 2)
        Y[0] = -np.inf
        Y[-1] = np.inf
        Y[1:-1] = y[idx_all_neighbours_and_semi_neighbours]
        idx_mem = {i: idx_all_neighbours_and_semi_neighbours[i-1] for i in range(1, Kprime+1)}
        sorted_indices = np.argsort(Y)[1:-1]
        # print(f'idx_mem: {idx_mem}')
        # print(f'idx_all_neighbours_and_semi_neighbours: {idx_all_neighbours_and_semi_neighbours}')
        # print(f'sorted_indices: {sorted_indices}')
        Y.sort()
        # print(f'Y: {Y}')

        # Line 4
        Alpha = np.array([(y[k_nearest.T[i]] <= y_i).sum() for i, y_i in enumerate(y)])
        # FIXME: Based on the description in ALRW, alpha_n = 0 initially, which seems to imply that 
        #        they consider -inf <= -inf to be false. Or possibly it is a consequence of assuming
        #        all labels and distances are distinct...
        N = np.array([(Alpha == k).sum() for k in range(self.k+1)])

        # Line 5
        L = -np.inf * np.ones(Kprime+1) # Initialize at something unreasonable
        U = -np.inf * np.ones(Kprime+1) # Initialize at something unreasonable
        L[0] = 0
        U[0] = N[0]/(n+1)

        # if Alpha[-1] > 0:
        #     print(f'n: {D.shape[0]}')
        #     print(f'Alpha: {Alpha}')
        #     print(f'N: {N}')
        #     print(f'y: {y}')
        #     print(f'{[k_nearest.T[i] for i, y_i in enumerate(y)]}')
        #     print(f'k_nearest: {k_nearest}')
            # print(f'Kprime: {Kprime}')

        # print(f'Kprime: {Kprime}')

        tic = time.time()
        # Line 6
        for k in range(1, Kprime+1):
            idx = idx_mem[sorted_indices[k-1]]
            # print(f'idx: {idx}')
            if (idx in full_neighbours + single_neighbours):
                # print(f'{idx} is a full or a single neighbour')
                N[Alpha[-1]] -= 1
                Alpha[-1] += 1
                N[Alpha[-1]] += 1
            if (idx in full_neighbours + semi_neighbours):
                # print(f'{idx} is a full or a semi-neighbour')
                N[Alpha[idx]] -= 1
                Alpha[idx] -= 1
                N[Alpha[idx]] += 1
            L[k] = N[:Alpha[-1]].sum() / (n+1) if Alpha[-1] != 0  else 0
            U[k] = N[:Alpha[-1] + 1].sum() / (n+1) if Alpha[-1] != 0  else N[0] / (n+1)
            # print(f'Alpha: {Alpha}')
            # print(f'Alpha_n: {Alpha[-1]}')
            # print(f'L[k]: {L[k]}')
            # print(f'N: {N}')
        # print(f'full_neighbours: {full_neighbours}')
        # print(f'single neighbours: {single_neighbours}')
        # print(f'semi_neighbours: {semi_neighbours}')
        toc_loop = time.time() - tic

        time_dict = {
            'Compute distance matrix': toc_dist,
            'Sort distance matrix': toc_sort,
            'Find all neighbours and semi-neighbours': toc_find_neighbours,
            'Loop': toc_loop
        }
        time_dict = time_dict if save_time else None
        # Line 12
        cpd = NearestNeighboursPredictiveDistributionFunction(L, U, Y, time_dict, epsilon=self.epsilon)

        if return_update:
            X = np.append(self.X, x.reshape(1, -1), axis=0)
            return cpd, {'X': X, 'D': D}
        else:
            return cpd
    

class DempsterHillConformalPredictiveSystem(ConformalPredictiveSystem):

    def __init__(self, verbose=0, rnd_state=None, epsilon=default_epsilon):
        '''
        The Dempster-Hill conformal predictive system uses only the labels of the examples, so the latter can be ignored.
        '''
        super().__init__(epsilon=epsilon)
        self.y = None
        self.verbose = verbose
        self.rnd_gen = np.random.default_rng(rnd_state)
    
    def learn_initial_training_set(self, y):
        self.y = y

    def learn_one(self, y):
        self.y = np.append(self.y, y)

    def predict_cpd(self, save_time=False):
        tic = time.time()
        Y = np.zeros(shape=self.y.shape[0] + 2)
        Y[0] = -np.inf
        Y[-1] = np.inf
        Y[1:-1] = self.y
        Y.sort()
        time_sort = time.time() - tic

        time_dict = {'Sort labels': time_sort} if save_time else None

        return DempsterHillConformalPredictiveDistribution(Y, time_dict, epsilon=self.epsilon)


class ConformalPredictiveDistributionFunction:

    '''
    NOTE
    The CPD contains all the information needed to form a
    prediction set. We can take quantiles and so on.
    '''

    def __init__(self, epsilon=default_epsilon):
        self.epsilon=epsilon

    def quantile(self, p, tau=None):
        raise NotImplementedError('Parent class has no quantile function')


    def predict_set(self, tau, epsilon=None, bounds='both', minimise_width=False):
        '''
        The convex hull of the epsilon/2 and 1-epsilon/2 quantiles make up
        the prediction set Gamma(epsilon)
        '''
        if epsilon is None:
            epsilon = self.epsilon

        if minimise_width:
            if bounds != 'both':
                raise Exception('bounds must be "both" if we are to minimise the interval width')
            def compute_width(delta, epsilon):
                epsilon_minus_delta = epsilon - delta
                return self.quantile(1-epsilon_minus_delta, tau) - self.quantile(delta, tau)

            delta = minimize_scalar(lambda x: compute_width(x, epsilon), bounds=(0, epsilon)).x
            eps_minus_delta = epsilon - delta
            lower = self.quantile(delta, tau)
            upper = self.quantile(1-eps_minus_delta, tau)

        
        else:
            q1 = epsilon/2
            q2 = 1 - epsilon/2
            if bounds=='both':
                lower = self.quantile(q1, tau)
                upper = self.quantile(q2, tau)
            elif bounds=='lower':
                lower = self.quantile(q1, tau)
                upper = np.inf
            elif bounds=='upper':
                lower = -np.inf
                upper = self.quantile(q2, tau)
            else: 
                raise Exception

        # print(f'Lower: {lower}')
        # print(f'Upper: {upper}')
        CP_int = get_ConformalPredictionInterval()
        return CP_int(lower, upper, epsilon)
    
    def find_smallest_epsilon(self, tau, increment=0.001):
        '''
        Find the smallest epsilon such that the prediction set is finite
        '''
        epsilon = 0
        while True:
            prediction_set = self.predict_set(tau=tau, epsilon=epsilon)
            if np.isfinite(prediction_set.width()):
                return epsilon
            epsilon += increment  # Increment epsilon by a small amount

    # These methods relate to when the cpd is used to predict sets
    @staticmethod
    def err(Gamma, y):
        return int(not(y in Gamma))
    

    @staticmethod
    def width(Gamma):
        return Gamma.width()
    

class RidgePredictiveDistributionFunction(ConformalPredictiveDistributionFunction):

    def __init__(self, C, time_dict=None, epsilon=default_epsilon):
        super().__init__(epsilon=epsilon)
        self.C = C
        self.L = np.array([self.__call__(y, 0) for y in self.C])
        self.U = np.array([self.__call__(y, 1) for y in self.C])
        self.Y = C

        self.time_dict = time_dict

        # What about
        self.y_vals = np.array(sorted([-np.inf, np.inf] + self.C[1: -1].tolist() + (self.C[1: -1] + MACHINE_EPSILON(self.C[1: -1])).tolist() + (self.C[1: -1] - MACHINE_EPSILON(self.C[1: -1])).tolist()))
        self.lowers = np.array([self.__call__(y, 0) for y in self.y_vals])
        self.uppers = np.array([self.__call__(y, 1) for y in self.y_vals])
        # Then the quantile can be computed by
        # self.y_vals[np.where((1 - tau) * self.L + tau * self.U >= p)[0].min()]

    def __call__(self, y, tau=None):
        if y == -np.inf:
            Pi0, Pi1 = 0.0, 0.0
        elif y == np.inf:
            Pi0, Pi1 = 1.0, 1.0
        else:
            C = self.C[:-1]
            idx_eq = np.where(y == C)[0]
            if idx_eq.shape[0] > 0:
                i_prime = idx_eq.min()
                i_bis = idx_eq.max()
                interval = ((i_prime - 1) / C.shape[0], (i_bis + 1) / C.shape[0])
            else:
                i = np.where(C <= y)[0].max()
                interval = (i / C.shape[0], (i + 1) / C.shape[0])

            Pi0 = interval[0]
            Pi1 = interval[1]            

        if tau is None:
            return Pi0, Pi1
        else:
            return (1 - tau) * Pi0 + tau * Pi1
        
    
    # NOTE: This takes forever if we have a large training set. 
    # Why not just invert?
    def quantile(self, p, tau=None):
        def compute_quantile(p, tau):
            # q = np.inf
            # y_vals = np.array(sorted([-np.inf, np.inf] + self.C[1: -1].tolist() + (self.C[1: -1] + MACHINE_EPSILON(self.C[1: -1])).tolist() + (self.C[1: -1] - MACHINE_EPSILON(self.C[1: -1])).tolist()))
            # # This loop is not very nice. Can we get rid of it?
            # for y in y_vals[::-1]:
            #     if self.__call__(y, tau) >= p:
            #         q = y
            #     else:
            #         return q
            # return q
            q = self.y_vals[np.where((1 - tau) * self.lowers + tau * self.uppers >= p)[0].min()]
            return q
        
        if tau is not None:
            q = compute_quantile(p, tau)
            return q
        else:
            q0 = compute_quantile(p, 0)
            q1 = compute_quantile(p, 1)
            return q0, q1
    
        
    def plot(self, tau=None):
        if tau is None:
            fig, ax = plt.subplots()
            ax.step(self.C, self.L, label=r'$\Pi(y, 0)$')
            ax.step(self.C, self.U, label=r'$\Pi(y, 1)$')
            ax.fill_between(self.C, self.L, self.U, step='pre', alpha=0.5, color='green')
            ax.set_ylabel('cumulative probability')
            ax.set_xlabel(r'$y$')
            ax.legend()
        else: 
            fig, ax = plt.subplots()
            ax.step(self.C, (1 - tau) * self.L + tau * self.U, label=r'$\Pi(y, \tau)$')
            ax.set_ylabel('cumulative probability')
            ax.set_xlabel(r'$y$')
            ax.legend()

        plt.close(fig)  # Prevent implicit display
        return fig


class NearestNeighboursPredictiveDistributionFunction(ConformalPredictiveDistributionFunction):
    '''
    TODO: Write tests
    '''
    # NOTE: It would be possible to pass a protection function here...
    def __init__(self, L, U, Y, time_dict=None, epsilon=default_epsilon):
        super().__init__(epsilon=epsilon)
        self.L = L 
        self.U = U
        self.Y = Y

        self.time_dict = time_dict
        
        Y_temp = Y[np.isfinite(Y)]
        self.y_vals = np.array(
            sorted(
                [-np.inf, np.inf] + 
                Y_temp.tolist() + 
                (Y_temp + MACHINE_EPSILON(Y_temp)).tolist() + 
                (Y_temp - MACHINE_EPSILON(Y_temp)).tolist())
        )
        self.lowers = np.array([self.__call__(y, 0) for y in self.y_vals])
        self.uppers = np.array([self.__call__(y, 1) for y in self.y_vals])

    def __call__(self, y, tau=None):
        # TODO: Check carefully that this is correct
        if y == self.Y[0]:
            Pi0, Pi1 = 0.0, 0.0
        elif y == self.Y[-1]:
            Pi0, Pi1 = 1.0, 1.0
        else:
            Y = self.Y[:-1]
            idx_eq = np.where(y == Y)[0]
            if idx_eq.shape[0] > 0:
                k = idx_eq.min()
                interval = (self.L[k-1], self.U[k])
            else:
                k = np.where(Y <= y)[0].max()
                interval = (self.L[k], self.U[k])
            
            Pi0 = interval[0]
            Pi1 = interval[1]            

        if tau is None:
            return Pi0, Pi1
        else:
            return (1 - tau) * Pi0 + tau * Pi1
        
    
    def quantile(self, p, tau=None):
        def compute_quantile(p, tau):
            # q = np.inf
            # y_vals = np.array(sorted([-np.inf, np.inf] + self.Y[1: -1].tolist() + (self.Y[1: -1] + MACHINE_EPSILON(self.Y[1: -1])).tolist() + (self.Y[1: -1] - MACHINE_EPSILON(self.Y[1: -1])).tolist()))
            # for y in y_vals[::-1]:
            #     if self.__call__(y, tau) >= p:
            #         q = y
            #     else:
            #         return q
            q = self.y_vals[np.where((1 - tau) * self.lowers + tau * self.uppers >= p)[0].min()]
            return q
        if tau is not None:
            q = compute_quantile(p, tau)
            return q
        else:
            q0 = compute_quantile(p, 0)
            q1 = compute_quantile(p, 1)
            return q0, q1
    
    def plot(self, tau=None):
        if tau is None:
            fig, ax = plt.subplots(sharex=True)
            ax.step(self.Y[1:], self.L, label=r'$\Pi(y, 0)$')
            ax.step(self.Y[1:], self.U, label=r'$\Pi(y, 1)$')
            ax.fill_between(self.Y[1:], self.L, self.U, step='pre', alpha=0.5, color='green')
            ax.legend()
        else: 
            fig, ax = plt.subplots()
            ax.step(self.Y[1:], (1 - tau) * self.L + tau * self.U, label=r'$\Pi(y, \tau)$')
            ax.legend()
        fig.tight_layout()
        plt.close(fig)  # Prevent implicit display
        return fig


class DempsterHillConformalPredictiveDistribution(ConformalPredictiveDistributionFunction):

    def __init__(self, Y, time_dict=None, epsilon=default_epsilon):
        super().__init__(epsilon=epsilon)
        self.Y = Y
        self.time_dict = time_dict

        self.y_vals = np.array(sorted([-np.inf, np.inf] + self.Y[1: -1].tolist() + (self.Y[1: -1] + MACHINE_EPSILON(self.Y[1: -1])).tolist() + (self.Y[1: -1] - MACHINE_EPSILON(self.Y[1: -1])).tolist()))
        self.lowers = np.array([self.__call__(y, 0) for y in self.y_vals])
        self.uppers = np.array([self.__call__(y, 1) for y in self.y_vals])

    
    def __call__(self, y, tau=None):
        if y == self.Y[0]:
            Pi0, Pi1 = 0.0, 0.0
        elif y == self.Y[-1]:
            Pi0, Pi1 = 1.0, 1.0
        else:
            Y = self.Y[:-1]
            idx_eq = np.where(y == Y)[0]
            if idx_eq.shape[0] > 0:
                k = idx_eq.min()
                interval = ((k-1)/(Y.shape[0]), (k+1)/(Y.shape[0]))
            else:
                k = np.where(Y <= y)[0].max()
                interval = ((k)/(Y.shape[0]), (k+1)/(Y.shape[0]))
            
            Pi0 = interval[0]
            Pi1 = interval[1]

        if tau is None:
            return Pi0, Pi1
        else:
            return (1 - tau) * Pi0 + tau * Pi1
        
    def quantile(self, p, tau=None):
        def compute_quantile(p, tau):
            # q = np.inf
            # y_vals = np.array(sorted([-np.inf, np.inf] + self.Y[1: -1].tolist() + (self.Y[1: -1] + MACHINE_EPSILON(self.Y[1: -1])).tolist() + (self.Y[1: -1] - MACHINE_EPSILON(self.Y[1: -1])).tolist()))
            # for y in y_vals[::-1]:
            #     if self.__call__(y, tau) >= p:
            #         q = y
            #     else:
            #         return q
            q = self.y_vals[np.where((1 - tau) * self.lowers + tau * self.uppers >= p)[0].min()]
            return q
        if tau is not None:
            q = compute_quantile(p, tau)
            return q
        else:
            q0 = compute_quantile(p, 0)
            q1 = compute_quantile(p, 1)
            return q0, q1

if __name__ == "__main__":
    import doctest
    import sys
    (failures, _) = doctest.testmod()
    if failures:
        sys.exit(1)
